"""
Module includes classes that accept single or multiple `arus.core.stream` instances, synchronize data from the streams, process the data using customized processor function, and output using the same iterator interface as `arus.core.stream`.

# Usage of `arus.core.pipeline.Pipeline`

# Single stream case

```python
.. include:: ../../examples/single_stream_pipeline.py
```

# Multiple streams case

```python
.. include:: ../../examples/multi_stream_pipeline.py
```

Author: Qu Tang

Date: 2019-11-15

License: see LICENSE file
"""

from .. import moment
import pathos.pools as ppools
import pathos.helpers as phelpers
import numpy as np
import queue
import pandas as pd
from loguru import logger
import threading
import time


class Pipeline:
    """The base class for a pipeline

    Pipeline class is a base class of any processor that accepts multiple `arus.core.stream` instances, sync and process them, and output the processed results.

    The base class provides the functionality to add, and sync multiple streams, as well as the functionality to add customized processor functions.

    Subclass may implement more complex logic such as saving the status of the processed results internally and reuse them with new data from the stream.

    Implementation details:
        Streams are served in threads. Data windows coming from the streams will be synced on a separate thread at first and the synced and merged data windows will be sent to process pools as process tasks for CPU intensive processing asynchronizedly. Another separate thread will wait for the completion of the process tasks and send the result to the result queue.

        This implementation, compared with previous version, makes the data processing most flexible considering that some processing needs to use data from more than one stream. Moreover, the customized process task can spawn other sub processes to further parallize the computational tasks for different combinations of stream data.

    Returns:
        pipeline (Pipeline): an instance object of type `Pipeline`.
    """

    def __init__(self, *, max_processes=None, scheduler='processes', preserve_status=False, name='default-pipeline'):
        """

        Args:
            max_processes (int, optional): the max number of sub processes to be spawned. Defaults to 'None'. If it is `None`, the max processes will be the number of cores - 2.
            scheduler (str, optional): scheduler to use, either 'processes' or 'threads'. Defaults to 'processes'.
            preserve_status (bool, optional): whether to preserve the previous input and output in the processor. If `True`, the second and third argument of processor function would be `prev_input` and `prev_output`. Defaults to `False`. When it is `True`, processors will run in sequence, meaning the next processor will start only when the previous one has completed.
            name (str, optional): the name of the pipeline. It will also be used as a prefix for all threads spawned by the class. Defaults to 'default-pipeline'.
        """
        self._started = False
        self._scheduler = scheduler
        self._max_processes = max_processes
        self._process_tasks = queue.Queue()
        self._queue = queue.Queue()
        self._prev_input = queue.Queue(1)
        self._prev_output = queue.Queue(1)
        self.name = name
        self._streams = []
        self._chunks = dict()
        self._stream_pointer = dict()
        self._streams_running = False
        self._process_cond = threading.Condition(threading.Lock())
        self._connected = False
        self._stop_sender = False
        self._preserve_status = preserve_status
        self._process_start_time = None
        self._pool = None

    @property
    def name(self):
        """`name` property getter

        Returns:
            str: the name of the pipeline
        """
        return self._name

    @name.setter
    def name(self, value):
        """`name` property setter

        Args:
            value (str): the name of the pipeline
        """
        self._name = value

    @property
    def started(self):
        """The status of the pipeline. Getter.

        Returns:
            bool: the status of the pipeline. 'True' means the pipeline is running.
        """
        return self._started

    @started.setter
    def started(self, value):
        """The status of the pipeline. Setter.

        Args:
            value (bool): the status of the pipeline. 'True' means the pipeline is running.
        """
        self._started = value

    def stop(self):
        return self.finish_tasks_and_stop()

    def finish_tasks_and_stop(self):
        """Gracefully shutdown the pipeline using the following procedure.

        1. It stops accepting incoming stream data.
        2. It allows sub-processes to finish existing tasks.
        3. It stops the sender thread.
        4. It stops the sub process pool to accept new tasks.
        5. It stops all incoming streams.
        6. It terminates the sub process pool.

        If any exceptions occur during the process, the function will return `False`.

        Returns:
            bool: `True` if the pipeline is shut down correctly.
        """
        try:
            logger.info('Stop pipeline...')
            self._connected = False
            time.sleep(0.2)
            if self._started:
                logger.info('Wait for processing tasks...')
                self._process_tasks.join()
                self._process_tasks.queue.clear()
                time.sleep(0.2)
            self._stop_sender = True
            with self._process_cond:
                self._started = False
                self._process_cond.notify_all()
            time.sleep(0.2)
            if self._pool is not None:
                logger.info('Close processing pool...')
                self._pool.close()
                self._pool.join()
            logger.info('Stop input streams...')
            for stream in self._streams:
                stream.stop()
            logger.info('Clear result queue...')
            with self._queue.mutex:
                self._queue.queue.clear()
            logger.info('Stopped.')
        except Exception as e:
            logger.error(e)
            return False
        return True

    def pause(self):
        """Pause processing the incoming streams, yet pipeline can still receive data from streams. Data will be ignored and not stored.
        """
        with self._process_cond:
            self._started = False
            self._process_start_time = None
        with self._queue.mutex:
            self._queue.queue.clear()

    def _is_running(self):
        if self.started:
            logger.warning(
                'It is not allowed to modify streams while pipeline is running, please stop the pipeline at first')
            return True
        else:
            return False

    def get_stream(self, stream_name):
        """Get the stream instance by its name

        Args:
            stream_name (str): The name of the stream

        Returns:
            arus.core.Stream: The instance of the stream if it is found, otherwise return `None`.
        """
        found = list(filter(lambda s: s._name == stream_name, self._streams))
        return None if len(found) == 0 else found[0]

    def add_stream(self, stream):
        """Add a new stream to the pipeline

        The stream will only be added when the pipeline is stopped.

        Args:
            stream (arus.core.Stream): an instance of arus.core.Stream class.
        """
        if self._is_running():
            return
        if self.get_stream(stream._name) is None:
            self._streams.append(stream)

    def set_processor(self, processor, **kwargs):
        """Set the processor function and its arguments

        The processor function should accept the following arg in the first place:
            chunk_list (list): A list of tuple including chunked data (one window) from all streams. Each tuple includes the following items.
            1. data: the chunked data of the stream
            2. name: the name of the stream

        Args:
            processor (object): A function to process chunks from the streams.
            kwargs (dict): keyword arguments passed to processor.
        """
        self._processor = processor
        self._processor_kwargs = kwargs

    def remove_stream(self, stream_name):
        """Remove stream from the pipeline

        Remove only works when the pipeline is stopped.

        Args:
            stream_name (str): The name of the stream.
        """
        if self._is_running():
            return
        found_stream = self.get_stream(stream_name)
        if found_stream is not None:
            self._streams.remove(found_stream)

    def _sync_streams(self):
        """
        Implementation details:
            This method depends on the stream to make sure it will not block for a long time when providing the chunk.

            This method also asssumes that when one window of data is missing, the stream should provide a notification, such as `None` for that window.

            Therefore, in the case when a Bluetooth device disconnects, the stream that serves the data of the device should always output `None` or empty DataFrame for those windows.

            In real time, because stream runs on separate thread, as long as the data is successfully passed to the stream queue, it won't block for a long time.
        """
        num_of_processors = min(phelpers.cpu_count() - 2, self._max_processes)
        if num_of_processors == 0:
            self._pool = None
        else:
            if self._pool is None:
                if self._scheduler == 'processes':
                    self._pool = ppools.ProcessPool(nodes=num_of_processors)
                    self._pool.restart(force=True)
                elif self._scheduler == 'threads':
                    self._pool = ppools.ThreadPool(nodes=num_of_processors)
                    self._pool.restart(force=True)
                else:
                    raise NotImplementedError(
                        'This scheduler is not supported: {}'.format(self._scheduler))
            else:
                self._pool.restart(force=True)
        while self._connected:
            if self._started:
                for stream in self._streams:
                    if stream.get_status() == stream.Status.RUN:
                        for data, context, name in stream.generate():
                            st = context['start_time']
                            et = context['stop_time']
                            prev_st = context['prev_start_time']
                            prev_et = context['prev_stop_time']
                            if self._is_data_after_start_time(st):
                                logger.debug(
                                    'recieved data window from ' + name)
                                self._chunks[st.timestamp()] = [] if st.timestamp(
                                ) not in self._chunks else self._chunks[st.timestamp()]
                                self._chunks[st.timestamp()].append(
                                    (data, st, et, prev_st, prev_et, name))
                                self._stream_pointer[name] = st.timestamp()
                            break
                        if self._is_data_after_start_time(st):
                            self._process_synced_chunks(
                                st, et, prev_st, prev_et, self.name)
                        else:
                            logger.debug('Discard one stream window' + str(
                                st) + 'coming before process start time: ' + str(self._process_start_time))
                if np.all([stream.get_status() != stream.Status.RUN for stream in self._streams]):
                    self._streams_running = False
                    break
            else:
                with self._process_cond:
                    # ignore the incoming stream data, the thread will stand by
                    self._process_cond.wait()

    def _send_result(self):
        while not self._stop_sender:
            if self._started:
                try:
                    task, st, et, prev_st, prev_et, name = self._process_tasks.get(
                        block=True, timeout=0.1)
                    if self._max_processes == 0:
                        result = task
                    else:
                        result = task.get()
                        self._process_tasks.task_done()
                    if self._preserve_status:
                        self._prev_output.put(result)
                    logger.info('Sending processed results to queue...')
                    self._put_result_in_queue(
                        (result, st, et, prev_st, prev_et, name))
                except queue.Empty:
                    if not self._streams_running:
                        self._put_result_in_queue(None)
                        break
            else:
                with self._process_cond:
                    # ignore if the pipeline is just connected but not started
                    self._process_cond.wait()

    def _process_synced_chunks(self, st, et, prev_st, prev_et, name):
        if st.timestamp() not in self._chunks:
            return
        chunk_list = self._chunks[st.timestamp()]
        if len(chunk_list) == len(self._streams):
            # this is the last stream chunk for st
            if self._preserve_status:
                if prev_st is None:
                    # the first window
                    prev_input = None
                    prev_output = None
                else:
                    prev_input = self._prev_input.get()
                    prev_output = self._prev_output.get()
                if self._max_processes == 0:
                    result = self._processor(
                        chunk_list, prev_input, prev_output, **self._processor_kwargs)
                    self._process_tasks.put(result)
                    del self._chunks[st.timestamp()]
                else:
                    try:
                        task = self._pool.apipe(
                            self._processor, chunk_list, prev_input, prev_output, **self._processor_kwargs)
                        self._process_tasks.put(
                            (task, st, et, prev_st, prev_et, name))
                        del self._chunks[st.timestamp()]
                    except ValueError as e:
                        logger.error(e)
                        return
                if self._preserve_status:
                    self._prev_input.put(chunk_list)
            else:
                if self._max_processes == 0:
                    result = self._processor(
                        chunk_list, **self._processor_kwargs)
                    self._process_tasks.put(result)
                    del self._chunks[st.timestamp()]
                elif self._pool is not None:
                    try:
                        logger.info('Starting a processing task...')
                        task = self._pool.apipe(self._processor, chunk_list,
                                                **self._processor_kwargs)
                        logger.info('Started a processing task...')
                        self._process_tasks.put(
                            (task, st, et, prev_st, prev_et, name))
                        del self._chunks[st.timestamp()]
                    except Exception as e:
                        logger.error(e)
                        return

    def _put_result_in_queue(self, result):
        self._queue.put(result)

    def _is_data_after_start_time(self, st):
        if self._process_start_time is not None:
            return st.timestamp() >= self._process_start_time.timestamp()
        else:
            return True

    def connect(self, start_time=None):
        """Connect and start the streams but not processing them.

        This function will start the streams one by one on daemon threads, and will start the syncing thread and sender thread as daemons in stand-by status.

        The streams will output data but data will be ignored by the pipeline until `Pipeline.process` got called.
        """
        self._started = False
        self._stop_sender = False
        self._connected = True
        self._sync_thread = threading.Thread(
            target=self._sync_streams, name=self._name + '-sync-streams')
        self._sync_thread.daemon = True
        self._sender_thread = threading.Thread(
            target=self._send_result, name=self._name + '-send-result')
        self._sender_thread.daemon = True
        self._sync_thread.start()
        self._sender_thread.start()
        for stream in self._streams:
            stream.start(start_time=start_time)
        self._streams_running = True
        return self

    def process(self, start_time=None):
        """Start processing the connected incoming stream data.

        Args:
            start_time (str or datetime or np.datetime64 or pd.Timestamp, optional): The start time to start accepting the incoming stream windows. If it is `None`, the pipeline will always output any incoming data without checking `start_time`.
        """
        self._process_start_time = start_time
        if self._process_start_time is not None:
            self._process_start_time = moment.Moment(
                self._process_start_time).to_pandas_timestamp()
        with self._process_cond:
            self._started = True
            self._process_cond.notify_all()
        return True

    def start(self, start_time=None, process_start_time=None):
        """Connect and process the incoming streams in a row together.

        Args:
            start_time (str or datetime or np.datetime64 or pd.Timestamp, optional): When multiple streams are provided, users must provide a valid start_time as a reference to sync between streams. If it is `None`, it will use the first timestamp from the stream data as start_time. Default is None. 
            process_start_time (str or datetime or np.datetime64 or pd.Timestamp, optional): Any stream windows coming before this time will be ignored. If it is `None`, nothing will be ignored. Defaults to None.
        """
        self.connect(start_time=start_time)
        self.process(start_time=process_start_time)

    def get_iterator(self, timeout=None):
        """Iterator for the processed results

        Raises:
            StopIteration: Raised when iterator ends

        Returns:
            Iteratable: an instance of a python iteratable for processed results
        """
        data_queue = self._queue

        class _result_iterator:
            def __iter__(self):
                return self

            def __next__(self):
                try:
                    data = data_queue.get(timeout=timeout)
                    if data is None:
                        # end of the stream, stop
                        raise StopIteration
                    return data
                except queue.Empty:
                    return None, None, None, None, None, None
        return _result_iterator()



import pandas as pd
import numpy as np
from .. import extensions as ext
import pytest
import itertools


def _as_2d_waterfall_signal(arg, repeat_n=3):
    return np.transpose(np.tile(arg, (repeat_n, 1)))


@pytest.fixture
def butterworth_test_signals():
    sr = 100
    t = np.linspace(0, 1, sr, False)
    mix_10_20_sig = np.sin(2*np.pi*10*t) + np.sin(2*np.pi*20*t)
    single_10_sig = np.sin(2*np.pi*10*t)
    constant_sig = np.ones(t.shape)
    zero_sig = np.zeros(t.shape)
    sigs = [mix_10_20_sig,
            _as_2d_waterfall_signal(mix_10_20_sig),
            constant_sig,
            _as_2d_waterfall_signal(constant_sig)
            ]

    low_pass_expects = [single_10_sig,
                        _as_2d_waterfall_signal(single_10_sig),
                        constant_sig,
                        _as_2d_waterfall_signal(constant_sig), ]

    low_pass_cases = list(
        itertools.starmap(
            lambda sig, expected: (
                {'X': sig, 'sr': sr, 'cut_offs': 15, 'order': 4,
                 'filter_type': 'low'},
                expected
            ),
            zip(sigs, low_pass_expects)
        )
    )

    band_pass_expects = [zero_sig,
                         _as_2d_waterfall_signal(zero_sig),
                         zero_sig,
                         _as_2d_waterfall_signal(zero_sig)]

    band_pass_cases = list(
        itertools.starmap(
            lambda sig, expected: (
                {'X': sig, 'sr': sr, 'cut_offs': [13, 17], 'order': 4,
                 'filter_type': 'pass'},
                expected),
            zip(sigs, band_pass_expects)
        )
    )

    return low_pass_cases + band_pass_cases


@pytest.fixture
def resample_test_signals():
    list_sr = [100, 80, 50, 30]
    test_cases = []
    signals = []
    for sr in list_sr:
        t = np.linspace([0, 0], [1, 1], sr, False, axis=0)
        mix_2_5_sig = np.sin(2*np.pi*2*t) + np.sin(2*np.pi*5*t)
        signals.append(mix_2_5_sig)
    sr_signal_pairs = zip(list_sr, signals)
    test_cases = itertools.combinations(sr_signal_pairs, 2)
    return test_cases


@pytest.fixture
def regularize_test_signals():
    results = []
    for et, expected in zip([1, 1.1, 1.11, 1.01], [80, 88, 88, 80]):
        t = np.linspace(0, et, num=85)
        X = np.random.rand(len(t), 3)
        results.append((t, X, 80, expected))
    return results


@pytest.fixture(scope='module')
def test_sensor_data(spades_lab_data):
    dw_sensor_file = spades_lab_data['subjects']['SPADES_1']['sensors']['DW'][0]
    da_sensor_file = spades_lab_data['subjects']['SPADES_1']['sensors']['DA'][0]
    dw_data = pd.read_csv(dw_sensor_file, parse_dates=[0])
    da_data = pd.read_csv(da_sensor_file, parse_dates=[0])
    st = max([dw_data.iloc[0, 0], da_data.iloc[0, 0]])
    dw_data = ext.pandas.segment_by_time(
        dw_data, st, st + pd.Timedelta(12.8 * 5, unit='seconds'))
    da_data = ext.pandas.segment_by_time(
        da_data, st, st + pd.Timedelta(12.8 * 5, unit='seconds'))
    return dw_data, da_data


class TestPandas:
    def test_merge_all(self):
        df1 = pd.DataFrame({'key': ['foo', 'bar', 'baz', 'foo'],
                            'value': [1, 2, 3, 5], 'group': [1, 1, 2, 2]})

        df2 = pd.DataFrame({'key': ['foo', 'bar', 'baz', 'foo'],
                            'value': [5, 6, 7, 8], 'group': [1, 1, 2, 2]})

        df3 = pd.DataFrame({'key': ['foo', 'bar', 'baz', 'foo'],
                            'value': [9, 10, 11, 12], 'group': [1, 1, 2, 2]})
        dfs = [df1, df2, df3]
        merged, cols_with_suffixes = ext.pandas.merge_all(*dfs, suffix_names=['DW', 'DA', 'DT'], suffix_cols=[
            'value'], on=['key', 'group'], how='inner', sort=False)

        np.testing.assert_array_equal(
            cols_with_suffixes, ['value_DW', 'value_DA', 'value_DT'])
        np.testing.assert_array_equal(
            merged[['key', 'group']], df1[['key', 'group']]
        )
        np.testing.assert_array_equal(
            set(merged.columns), set(['key', 'value_DW', 'value_DA', 'value_DT', 'group']))

        merged, cols_with_suffixes = ext.pandas.merge_all(df1, suffix_names=['DW'], suffix_cols=[
            'value'], on=['key', 'group'], how='inner', sort=False)
        np.testing.assert_array_equal(merged.values, df1.values)
        np.testing.assert_array_equal(
            set(merged.columns), set(['key', 'value_DW', 'group']))

    def test_filter_column(self):
        df1 = pd.DataFrame({'key': ['foo', 'bar', 'baz', 'foo'],
                            'value': [1, 2, 3, 5], 'group': [1, 1, 2, 2]})
        filtered = ext.pandas.filter_column(
            df1, col='key', values_to_filter_out=['foo'])
        np.testing.assert_array_equal(filtered['value'].values, [2, 3])
        np.testing.assert_array_equal(filtered['key'].values, ['bar', 'baz'])

        filtered = ext.pandas.filter_column(
            df1, col='value', values_to_filter_out=[2, 5])
        np.testing.assert_array_equal(filtered['value'].values, [1, 3])
        np.testing.assert_array_equal(filtered['key'].values, ['foo', 'baz'])

    @pytest.mark.parametrize('start_time', [None, "2015-09-24 14:17:00.000"])
    @pytest.mark.parametrize('stop_time', [None, "2015-09-24 14:17:00.000"])
    def test_get_common_timespan(self, test_sensor_data, start_time, stop_time):
        dw_data, da_data = test_sensor_data
        st, et = ext.pandas.get_common_timespan(
            dw_data, da_data, st=start_time, et=stop_time)
        if start_time is None:
            assert st.timestamp() == pd.Timestamp(
                '2015-09-24 14:17:48.013').timestamp()
        else:
            assert st.timestamp() == pd.Timestamp(start_time).timestamp()

        if stop_time is None:
            assert et.timestamp() == pd.Timestamp(
                '2015-09-24 14:18:52.000').timestamp()
        else:
            assert et.timestamp() == pd.Timestamp(stop_time).timestamp()

    @pytest.mark.parametrize('start_time', [None, "2015-09-24 14:17:48.000"])
    @pytest.mark.parametrize('stop_time', [None, "2015-09-24 14:17:00.000"])
    def test_split_into_windows(self, test_sensor_data, start_time, stop_time):
        dw_data, da_data = test_sensor_data
        window_start_markers = ext.pandas.split_into_windows(
            dw_data, da_data, step_size=12.8, st=start_time, et=stop_time)
        if start_time is None and stop_time is None:
            assert len(window_start_markers) == 5
            assert np.all(
                np.diff(window_start_markers)[:-1].astype('timedelta64[ms]').astype(int) == 12800)
            assert window_start_markers[0].timestamp() == pd.Timestamp(
                '2015-09-24 14:17:48.013').timestamp()
        elif start_time is not None and stop_time is None:
            assert len(window_start_markers) == 5
            assert np.all(
                np.diff(window_start_markers)[:-1].astype('timedelta64[ms]').astype(int) == 12800)
            assert window_start_markers[0].timestamp() == pd.Timestamp(
                start_time).timestamp()
        elif stop_time is not None:
            assert len(window_start_markers) == 0

    def test_fixed_window_slider(self, test_sensor_data):
        def sample_counter(*dfs, st, et, placements):
            result = {
                'START_TIME': [st],
                'STOP_TIME': [et]
            }
            for df, p in zip(dfs, placements):
                result[p] = [df.shape[0]]
            return pd.DataFrame.from_dict(result)

        dw_df, da_df = test_sensor_data
        count_df = ext.pandas.fixed_window_slider(
            *[dw_df, da_df], slider_fn=sample_counter, window_size=12.8, step_size=None, placements=['DW', 'DA'])
        assert count_df.shape[1] == 4
        assert count_df.shape[0] == 5
        assert (count_df.iloc[:, 2] == 1024).all()


class TestNumpy:
    def test_mutate_nan(self):
        # test signal without nan
        X = np.random.rand(10, 3)
        X_new = ext.numpy.mutate_nan(X)
        np.testing.assert_array_almost_equal(X, X_new)

        X = np.random.rand(10, 1)
        X_new = ext.numpy.mutate_nan(X)
        np.testing.assert_array_almost_equal(X, X_new)

        # test signal with single sample without nan
        X = np.array([[0.]])
        X_new = ext.numpy.mutate_nan(X)
        np.testing.assert_array_almost_equal(X, X_new)

        X = np.array([[0., 0, 0, ]])
        X_new = ext.numpy.mutate_nan(X)
        np.testing.assert_array_almost_equal(X, X_new)

        # test signal with nan
        X = np.atleast_2d(np.sin(2*np.pi * 1 * np.arange(0, 1, 1.0 / 100))).T
        X_nan = np.copy(X)
        X_nan[5:10, 0] = np.nan
        X_new = ext.numpy.mutate_nan(X_nan)
        np.testing.assert_array_almost_equal(X, X_new, decimal=4)

        X = np.tile(np.sin(2*np.pi * 1 *
                           np.arange(0, 1, 1.0 / 100.)), (3, 1)).T
        X_nan = np.copy(X)
        X_nan[5:10, 0:3] = np.nan
        X_new = ext.numpy.mutate_nan(X_nan)
        np.testing.assert_array_almost_equal(X, X_new, decimal=4)

    def test_butterworth(self, butterworth_test_signals):
        for args, expected in butterworth_test_signals:
            result = ext.numpy.butterworth(**args)
            diff = expected[10:90, ] - result[10:90, ]
            assert expected.shape == result.shape
            assert np.all(diff < 0.1)

    def test_resample(self, resample_test_signals):
        for test_signal_pair_1, test_signal_pair_2 in resample_test_signals:
            sr1 = test_signal_pair_1[0]
            signal1 = test_signal_pair_1[1]
            sr2 = test_signal_pair_2[0]
            signal2 = test_signal_pair_2[1]
            result12 = ext.numpy.resample(signal1, sr1, sr2)
            expected12 = signal2
            st = int(sr2/10)
            et = int(sr2 - sr2/10)
            diff12 = expected12[st:et, ] - \
                result12[st:et, ]
            print('{}, {}'.format(sr1, sr2))
            assert expected12.shape == result12.shape
            assert np.all(diff12 < 0.1)

            result21 = ext.numpy.resample(signal2, sr2, sr1)
            expected21 = signal1
            st = int(sr1/10)
            et = int(sr1 - sr1/10)
            diff21 = expected21[st:et, ] - \
                result21[st:et, ]
            print('{}, {}'.format(sr2, sr1))
            assert expected21.shape == result21.shape
            assert np.all(diff21 < 0.1)

    def test_apply_over_subwins(self):
        func = np.mean
        # test on single row array with subwins and subwin_samples not set
        X = ext.numpy.atleast_float_2d(np.array([1., 1., 1.]))
        result = ext.numpy.apply_over_subwins(
            X, func, subwin_samples=None, subwins=None, axis=0)
        assert np.array_equal(result, X)

        # test on single row array with subwin_samples not set
        X = ext.numpy.atleast_float_2d(np.array([1., 1., 1.]))
        result = ext.numpy.apply_over_subwins(
            X, func, subwin_samples=None, subwins=1, axis=0)
        assert np.array_equal(result, X)

        # test on single row array with subwins not set
        X = ext.numpy.atleast_float_2d(np.array([1., 1., 1.]))
        result = ext.numpy.apply_over_subwins(
            X, func, subwin_samples=1, subwins=None, axis=0)
        assert np.array_equal(result, X)

        # test on single row array with subwins to be zero
        X = ext.numpy.atleast_float_2d(np.array([1., 1., 1.]))
        result = ext.numpy.apply_over_subwins(
            X, func, subwin_samples=2, subwins=None, axis=0)
        assert np.array_equal(result, X)
        result = ext.numpy.apply_over_subwins(
            X, func, subwin_samples=None, subwins=0, axis=0)
        assert np.array_equal(result, X)

        # test on single row array with subwin_samples to be zero
        X = ext.numpy.atleast_float_2d(np.array([1., 1., 1.]))
        result = ext.numpy.apply_over_subwins(
            X, func, subwin_samples=0, subwins=None, axis=0)
        assert np.array_equal(result, X)
        result = ext.numpy.apply_over_subwins(
            X, func, subwin_samples=None, subwins=2, axis=0)
        assert np.array_equal(result, X)

        # test on 2d array
        X = ext.numpy.atleast_float_2d(np.ones((10, 3)))
        result = ext.numpy.apply_over_subwins(
            X, func, subwin_samples=2, subwins=None, axis=0)
        assert np.array_equal(result, np.ones((5, 3)))
        result = ext.numpy.apply_over_subwins(
            X, func, subwin_samples=None, subwins=2, axis=0)
        assert np.array_equal(result, np.ones((2, 3)))

        # test on 2d array use subwins at first
        X = ext.numpy.atleast_float_2d(np.ones((10, 3)))
        result = ext.numpy.apply_over_subwins(
            X, func, subwin_samples=2, subwins=2, axis=0)
        assert np.array_equal(result, np.ones((2, 3)))

        # test on 2d array use window parameters that are not fully dividable
        X = ext.numpy.atleast_float_2d(
            np.concatenate((
                np.ones((1, 3)) * 2,
                np.ones((8, 3)),
                np.ones((1, 3)) * 2),
                axis=0)
        )
        result = ext.numpy.apply_over_subwins(
            X, func, subwin_samples=4, subwins=None, axis=0)
        assert np.array_equal(result, np.ones((2, 3)))

        # test on 2d array use window parameters that are not fully dividable
        X = ext.numpy.atleast_float_2d(
            np.concatenate((
                np.ones((1, 3)) * 2,
                np.ones((3, 3)),
                np.ones((6, 3)) * 2),
                axis=0)
        )
        result = ext.numpy.apply_over_subwins(
            X, func, subwin_samples=3, subwins=None, axis=0)
        assert np.array_equal(result, np.array(
            [[1, 1, 1], [2, 2, 2], [2, 2, 2]]))

    def test_regularize_sr(self, regularize_test_signals):
        for t, X, sr, expected in regularize_test_signals:
            new_t, new_X = ext.numpy.regularize_sr(t, X, sr)
            assert len(new_t) == expected
            np.testing.assert_array_equal(new_X.shape, [expected, 3])

    def test_vector_magnitude(self):
        # test with a single row of data
        X = np.array([[1., 1., 1.]])
        result = ext.numpy.vector_magnitude(X)
        assert np.allclose(result, np.sqrt(3), atol=0.001)
        # test with an array
        X = np.array([[1., 1., 1.], [1., 1., 1.]])
        result = ext.numpy.vector_magnitude(X)
        assert np.allclose(result, np.array(
            [[np.sqrt(3)], [np.sqrt(3)]]), atol=0.001)
        # test with NaN data
        X = np.array([[1., np.nan, 1.]])
        result = ext.numpy.vector_magnitude(X)
        assert np.isnan(result)

    def test_flip_and_swap(self):
        # test on 1d data
        # test with flip
        X = np.array([[1., 1., 1.]])
        result = ext.numpy.flip_and_swap(
            X, x_flip='x', y_flip='y', z_flip='-z')
        assert np.allclose(result, np.array([[1., 1., -1.]]), atol=0.001)
        # test with swap
        X = np.array([[1., 2., 3.]])
        result = ext.numpy.flip_and_swap(X, x_flip='y', y_flip='x', z_flip='z')
        assert np.allclose(result, np.array([[2., 1., 3.]]), atol=0.001)
        # test with flip and swap
        X = np.array([[1., 2., 3.]])
        result = ext.numpy.flip_and_swap(
            X, x_flip='y', y_flip='-x', z_flip='z')
        assert np.allclose(result, np.array([[2., -1., 3.]]), atol=0.001)
        # test on 2d data
        X = np.array([[1., 1., 1.], [1, 2, 3]])
        result = ext.numpy.flip_and_swap(
            X, x_flip='y', y_flip='-x', z_flip='z')
        assert np.allclose(result, np.array(
            [[1., -1., 1.], [2., -1., 3.]]), atol=0.001)




# %% [markdown]
# ## Import dependencies

# %%
import arus
from loguru import logger

# %% [markdown]
# ## stream with random data generator

# %%
arus.dev.set_default_logger()

gr = arus.generator.RandomAccelDataGenerator(
    50, buffer_size=25, max_samples=75)
seg = arus.segmentor.SlidingWindowSegmentor(window_size=1)

stream = arus.stream2.Stream(
    generator=gr, segmentor=seg, name='random-accel-stream')

stream.start()

i = 0
for data in stream.get_result():
    if data.signal == arus.Node.Signal.DATA:
        logger.info(data)
        if i == 3 or data.values is None:
            break
        i += 1

stream.stop()

# %% [markdown]
# ## stream with metawear generator

# %%
arus.dev.set_default_logger()

addrs = arus.plugins.metawear.MetaWearScanner().get_nearby_devices(max_devices=1)
gr = arus.plugins.metawear.MetaWearAccelDataGenerator(
    addrs[0], sr=50, grange=8, buffer_size=25)
seg = arus.segmentor.SlidingWindowSegmentor(window_size=1)

stream = arus.stream2.Stream(
    generator=gr, segmentor=seg, name='metawear-stream')

stream.start()

i = 0
for data in stream.get_result():
    if data.signal == arus.Node.Signal.DATA:
        logger.info(data)
        if i == 3 or data.values is None:
            break
        i += 1

stream.stop()



from .. import mhealth_format as mh
from .. import generator, plugins, segmentor, stream2
from .. import class_label
from dataclasses import dataclass, field
from loguru import logger
import pandas as pd
import typing
import os
import enum
import re


class InputType(enum.Enum):
    MHEALTH_FORMAT = "MHEALTH_FORMAT"
    SIGNALIGNER = "SIGNALIGNER"


@dataclass
class SensorObj:
    paths: str
    sid: str = None
    data_type: str = None
    device_type: str = None
    placement: str = None
    sr: int = None
    grange: float = None
    data: pd.DataFrame = None
    start_time: typing.Any = None
    input_type: InputType = InputType.MHEALTH_FORMAT

    def __post_init__(self):
        if self.input_type == InputType.SIGNALIGNER:
            reader = plugins.actigraph.ActigraphReader(
                self.paths[0], has_ts=False, has_header=True)
            reader.read_meta()
            meta = reader.get_meta()
            self.start_time = meta['START_TIME']
            self.sr = meta['SAMPLING_RATE']
            self.grange = meta['DYNAMIC_RANGE']
            self.sid = meta['SENSOR_ID']
            self.data_type = os.path.basename(self.paths[0]).split('.csv')[0]
        elif self.input_type == InputType.MHEALTH_FORMAT:
            self.sid = mh.parse_sensor_id_from_filepath(self.paths[0])
            self.data_type = mh.parse_data_type_from_filepath(self.paths[0])
            self.device_type = mh.parse_sensor_type_from_filepath(
                self.paths[0])
            if self.device_type == 'ActigraphGT9X':
                if self.data_type == 'AccelerationCalibrated' or self.data_type == 'AccelerometerCalibrated':
                    self.sr = 80
                elif self.data_type == 'IMUMagnetometer' or self.data_type == 'IMUAccelerometerCalibrated' or self.data_type == 'IMUGyroscope':
                    self.sr = 100
                else:
                    self.sr = None
            self.start_time = mh.parse_timestamp_from_filepath_content(
                self.paths[0])

    def get_data(self):
        if self.input_type == InputType.MHEALTH_FORMAT:
            return mh.MhealthFileReader.read_csvs(*self.paths)
        elif self.input_type == InputType.SIGNALIGNER:
            reader = plugins.actigraph.ActigraphReader(
                self.paths[0], False, False)
            reader.read()
            return next(reader.get_data())

    def get_stream(self, seg: segmentor.Segmentor = None, stream_id: str = None, **seg_kwargs):
        if self.input_type == InputType.MHEALTH_FORMAT:
            gen = generator.MhealthSensorFileGenerator(
                *self.paths, buffer_size=3600 * 80)
        elif self.input_type == InputType.SIGNALIGNER:
            gen = plugins.actigraph.ActigraphSensorFileGenerator(
                *self.paths, has_ts=False, has_header=True, buffer_size=3600 * 80)
        else:
            logger.error(
                f'Unrecognized input type: {self.input_type.name}, failed to get stream')
            return None
        seg = seg or segmentor.SlidingWindowSegmentor(**seg_kwargs)
        stream_name = self.sid + '-' + self.data_type + '-stream'
        sensor_stream = stream2.Stream(gen, seg, name=stream_name)
        sensor_stream.set_essential_context(
            stream_id=stream_id or self.placement or stream_name)
        return sensor_stream


@dataclass
class AnnotationObj:
    paths: str
    aid: str = None
    annotation_type: str = None
    annotator: str = None
    data: pd.DataFrame = None
    input_type: InputType = InputType.MHEALTH_FORMAT

    def __post_init__(self):
        if self.input_type == InputType.MHEALTH_FORMAT:
            self.annotation_type = mh.parse_annotation_type_from_filepath(
                self.paths[0])
            self.annotator = mh.parse_annotator_from_filepath(self.paths[0])
            self.aid = mh.parse_data_id_from_filepath(self.paths[0])

    def get_stream(self, window_size: float, start_time: "str, pandas.Timestamp", input_type: InputType):
        if input_type == InputType.MHEALTH_FORMAT:
            gen = generator.MhealthSensorFileGenerator(
                *self.paths, buffer_size=3600 * 80)
        elif input_type == InputType.SIGNALIGNER:
            gen = plugins.actigraph.ActigraphSensorFileGenerator(
                *self.paths, buffer_size=3600 * 80)
        else:
            logger.error(
                f'Unrecognized input type: {input_type.name}, failed to get stream')
            return None
        seg = segmentor.SlidingWindowSegmentor(
            window_size=window_size, ref_st=start_time, st_col=0, et_col=None)
        sensor_stream = stream2.Stream(gen, seg, name=self.sid + '-' +
                                       self.data_type + '-stream')
        return sensor_stream


@dataclass
class SubjectObj:
    path: str
    pid: str
    demography: pd.DataFrame = None
    sensors: typing.List[SensorObj] = field(default_factory=list)
    annotations: typing.List[AnnotationObj] = field(default_factory=list)
    processed: typing.Dict[str, typing.Any] = field(default_factory=dict)

    def get_sensor(self, field_name, field_value):
        for sensor in sensors:
            if sensor.__getattribute__(field_name) == field_value:
                return sensor
        return None

    def subset_sensors(self, field_name, field_value):
        sensors = []
        for sensor in self.sensors:
            if re.fullmatch(field_value, sensor.__getattribute__(field_name)) is not None:
                sensors.append(sensor)
        return sensors


@dataclass
class MHDataset:
    name: str
    path: str = None
    load: bool = True
    input_type: InputType = InputType.MHEALTH_FORMAT
    subjects: typing.List[SubjectObj] = field(default_factory=list)
    processed: typing.Dict[str, typing.Any] = field(default_factory=dict)
    class_set_parser: object = None
    placement_parser: object = None

    def __post_init__(self):
        if not self.load:
            return
        self._load()

    def _load(self):
        self.subjects = []
        if self.path is None:
            return
        pids = mh.get_pids(self.path)
        for pid in pids:
            subj = SubjectObj(path=os.path.join(self.path, pid),
                              pid=pid)
            subj.sensors = self._load_sensors(pid)
            subj.annotations = self._load_annotations(pid)
            self.subjects.append(subj)

    def _load_sensors(self, pid):
        sensor_files = mh.get_sensor_files(pid, self.path)
        data_ids = sorted(list(set(
            map(lambda f: mh.parse_data_id_from_filepath(f), sensor_files))))
        sensors = []
        for data_id in data_ids:
            paths = list(filter(lambda f: data_id in f, sensor_files))
            sid = mh.parse_sensor_id_from_filepath(paths[0])
            if self.placement_parser is not None:
                placement = self.placement_parser(self.path, pid, sid)
            else:
                placement = None
            sensor = SensorObj(paths=paths, sid=sid, placement=placement)
            sensors.append(sensor)
        return sensors

    def _load_annotations(self, pid):
        annotation_files = mh.get_annotation_files(pid, self.path)
        annotations = []
        data_ids = sorted(list(set(
            map(lambda f: mh.parse_data_id_from_filepath(f), annotation_files))))
        for data_id in data_ids:
            paths = list(filter(lambda f: data_id ==
                                mh.parse_data_id_from_filepath(f), annotation_files))
            annot = AnnotationObj(
                paths=paths, aid=data_id)
            annotations.append(annot)
        return annotations

    def clear_processed(self):
        self.processed = {}
        for subj in self.subjects:
            subj.processed = {}

    def set_class_set_parser(self, class_set_parser):
        self.class_set_parser = class_set_parser

    def set_placement_parser(self, placement_parser=None):
        self.placement_parser = placement_parser
        if placement_parser is not None:
            logger.info('Set placement func, loading placements...')
            for pid in self.get_pids():
                for sensor in self.get_sensors(pid):
                    # logger.debug(f'Set placement for {pid} - {sensor.sid}')
                    sensor.placement = placement_parser(
                        self.path, pid, sensor.sid)

    def has_pid(self, pid):
        for subj in self.subjects:
            if pid == subj.pid:
                return True
        return False

    def get_pids(self):
        return list(map(lambda subj: subj.pid, self.subjects))

    def get_subject_obj(self, pid):
        for subj in self.subjects:
            if pid == subj.pid:
                return subj
        return None

    def get_sensors(self, pid, sid=None, placement=None, data_type=None, device_type=None):
        subj = self.get_subject_obj(pid)

        def filter_func(s):
            if sid is not None and s.sid is not None and sid != s.sid:
                return False
            if placement is not None and s.placement is not None and placement != s.placement:
                return False
            if data_type is not None and s.data_type is not None and data_type != s.data_type:
                return False
            if device_type is not None and s.device_type is not None and device_type != s.device_type:
                return False
            return True
        if subj is not None:
            sensors = subj.sensors
            sensor_objs = list(filter(filter_func, sensors))
            return sensor_objs
        else:
            logger.error(f'{pid} does not exist.')
            return []

    def get_annotations(self, pid, annotation_type=None, annotator=None, aid=None):
        subj = self.get_subject_obj(pid)

        def filter_func(s):
            if annotation_type is not None and s.annotation_type is not None and annotation_type != s.annotation_type:
                return False
            if annotator is not None and s.annotator is not None and annotator != s.annotator:
                return False
            if aid is not None and s.aid is not None and aid != s.aid:
                return False
            return True
        if subj is not None:
            annots = subj.annotations
            annot_objs = list(filter(filter_func, annots))
            return annot_objs
        else:
            logger.error(f'{pid} does not exist.')
            return []

    def get_annotation_field(self, pid, field_name):
        subj = self.get_subject_obj(pid)
        if subj is not None:
            annots = subj.annotations
            field_values = set(map(
                lambda a: a.__getattribute__(field_name), annots))
            return field_values
        else:
            logger.error(f'{pid} does not exist.')
            return []

    def get_sensor_field(self, pid, field_name):
        subj = self.get_subject_obj(pid)
        if subj is not None:
            sensors = subj.sensors
            field_values = set(map(
                lambda s: s.__getattribute__(field_name), sensors))
            return field_values
        else:
            logger.error(f'{pid} does not exist.')
            return []

    def get_class_set(self, pid, task_names, window_size, step_size=None, start_time=None, stop_time=None, show_progress=True):
        step_size = step_size or window_size
        aids = self.get_annotation_field(pid, 'aid')
        raw_resources = [mh.MhealthFileReader.read_csvs(
            *self.get_annotations(pid, aid=aid)[0].paths, datetime_cols=[0, 1, 2]) for aid in aids]
        class_set = class_label.ClassSet(raw_resources, aids)
        class_set.compute_offline(window_size, self.class_set_parser,
                                  task_names, start_time, stop_time, step_size=step_size, pid=pid, show_progress=show_progress)
        return class_set.get_class_set()

    def subset(self, name, pids=None):
        new_ds = MHDataset(path=self.path, name=name,
                           input_type=self.input_type, load=False)
        if pids is None:
            pids = self.get_pids()
        for pid in pids:
            new_ds.subjects.append(self.get_subject_obj(pid))
        new_ds.set_class_set_parser(self.class_set_parser)
        new_ds.set_placement_parser(self.placement_parser)
        return new_ds

    def get_session_span(self, pid):
        if self.input_type == InputType.MHEALTH_FORMAT:
            return mh.get_session_span(pid, self.path)
        else:
            logger.error(
                f'Unrecognized dataset input type {self.input_type.name}')
            return None


@dataclass
class SGDataset(MHDataset):
    name: str
    path: str = None
    sensor_folder: str = None
    label_folder: str = None
    load: bool = True
    input_type: InputType = InputType.SIGNALIGNER
    subjects: typing.List[SubjectObj] = field(default_factory=list)
    processed: typing.Dict[str, typing.Any] = field(default_factory=dict)
    class_set_parser: object = None
    placement_parser: object = None

    def _load(self):
        self.subjects = []
        pid = self.name
        subj = SubjectObj(path=self.path, pid=pid)
        subj.sensors = self._load_sensors(pid)
        subj.annotations = self._load_annotations(pid)
        self.subjects.append(subj)

    def _load_sensors(self, pid):
        sensors = []
        if self.sensor_folder is None:
            logger.error('sensor_folder must be set to load sensor files.')
            return sensors
        path = os.path.join(self.path, self.sensor_folder)
        sensor_filenames = sorted(os.listdir(path))

        for sensor_filename in sensor_filenames:
            data_id = sensor_filename.split('.csv')[0]
            paths = [os.path.join(path, sensor_filename)]
            sid = data_id
            if self.placement_parser is not None:
                placement = self.placement_parser(self.path, pid, sid)
            else:
                placement = None
            data_type = data_id
            device_type = data_id
            sensor = SensorObj(paths=paths, sid=sid, data_type=data_type,
                               device_type=device_type, placement=placement, input_type=self.input_type)
            sensors.append(sensor)
        return sensors

    def _load_annotations(self, pid):
        annotations = []
        if self.label_folder is None:
            logger.error(
                'label_folder must be set to load annotation/label files.')
            return annotations
        path = os.path.join(self.path, self.label_folder)
        label_filenames = sorted(os.listdir(path))

        for label_filename in label_filenames:
            data_id = label_filename.split('.csv')[0]
            paths = [data_id]
            annotation_type = data_id
            annotator = data_id
            annot = AnnotationObj(
                paths=paths, aid=data_id, annotation_type=annotation_type, annotator=annotator, input_type=self.input_type)
            annotations.append(annot)
        return annotations



import numpy as np
import pytest

from .. import dataset as ds
from .. import feature_vector as fv
from .. import mhealth_format as mh


@pytest.fixture
def test_data(single_mhealth_sensor):
    data = single_mhealth_sensor.iloc[:1000, :]
    st = data.iloc[0, 0]
    et = data.iloc[-1, 0]
    sr = 80
    return data, st, et, sr


@ pytest.mark.parametrize('use_vm', [True, False])
def test_inertial_single_triaxial(test_data, use_vm):

    raw_df, st, et, sr = test_data

    fv_df, fv_names = fv.inertial.single_triaxial(
        raw_df, sr=sr, st=st, et=et, subwin_secs=2, ori_unit='rad', activation_threshold=0.2, use_vm=use_vm)

    feature_names = fv.inertial.assemble_fv_names(use_vm=use_vm)
    np.testing.assert_array_equal(fv_names, feature_names)

    assert fv_df.shape[0] == 1
    assert np.sum(fv_df.notna().values) + \
        (2 if use_vm else 3) == fv_df.shape[1]



"""
Demonstration of the usage of arus.models.muss.get_inference_pipeline
=====================================================================

The pipeline uses multiple metawear streams.
"""

from arus.models.muss import MUSSModel
from arus.testing import load_test_data
import pandas as pd
import arus
from datetime import datetime


def train_test_classifier(muss):
    feature_filepath, _ = load_test_data(file_type='mhealth', sensor_type='feature',
                                         file_num='single', exception_type='multi_placements')
    class_filepath, _ = load_test_data(file_type='mhealth', sensor_type='class_labels',
                                       file_num='single', exception_type='multi_tasks')
    feature_set = pd.read_csv(
        feature_filepath, infer_datetime_format=True, parse_dates=[0, 1, 2])
    dw_feature_set = feature_set.loc[feature_set['SENSOR_PLACEMENT'] == 'DW', [
        'HEADER_TIME_STAMP', 'START_TIME', 'STOP_TIME'] + muss.get_feature_names()]
    da_feature_set = feature_set.loc[feature_set['SENSOR_PLACEMENT'] == 'DA', [
        'HEADER_TIME_STAMP', 'START_TIME', 'STOP_TIME'] + muss.get_feature_names()]
    combined_feature_set, feature_names = muss.combine_features(
        dw_feature_set, da_feature_set, placement_names=['DW', 'DA'], group_col=None)
    class_set = pd.read_csv(
        class_filepath, infer_datetime_format=True, parse_dates=[0, 1, 2])
    class_set = class_set[['HEADER_TIME_STAMP',
                           'START_TIME', 'STOP_TIME', 'MUSS_3_POSTURES']]
    input_feature, input_class = muss.sync_feature_and_class(
        combined_feature_set, class_set)
    filtered_feature, filtered_class = muss.remove_classes(
        input_feature, input_class, class_col='MUSS_3_POSTURES', classes_to_remove=['Unknown', 'Transition'])
    model = muss.train_classifier(
        filtered_feature, filtered_class, class_col='MUSS_3_POSTURES', feature_names=feature_names, placement_names=['DW', 'DA'])
    return model


def prepare_streams():
    generator = arus.plugins.metawear.MetaWearAccelDataGenerator(
        "D2:C6:AF:2B:DB:22", sr=50, grange=8, buffer_size=100)
    segmentor = arus.segmentor.SlidingWindowSegmentor(window_size=12.8)
    stream1 = arus.Stream(generator, segmentor, name='DW')
    generator = arus.plugins.metawear.MetaWearAccelDataGenerator(
        "FF:EE:B8:99:0C:64", sr=50, grange=8, buffer_size=100)
    segmentor = arus.segmentor.SlidingWindowSegmentor(window_size=12.8)
    stream2 = arus.Stream(generator, segmentor, name='DA')
    return stream1, stream2


if __name__ == "__main__":
    muss = MUSSModel()
    model = train_test_classifier(muss)
    stream1, stream2 = prepare_streams()
    kwargs = {
        'DW': {
            'sr': 50
        },
        'DA': {
            'sr': 100
        },
        'model': model,
        'scheduler': 'processes',
        'max_processes': 2
    }
    start_time = datetime.now()
    muss_pipeline = muss.get_inference_pipeline(
        stream1, stream2, name='muss-pipeline', **kwargs)
    muss_pipeline.start(start_time=start_time, process_start_time=start_time)
    i = 0
    for data, st, et, _, _, name in muss_pipeline.get_iterator(timeout=0.2):
        if data is None:
            continue
        i = i + 1
        print(str(datetime.now()) + ': ' + str(st) + ' - ' + str(et))
        print(model[0].classes_)
        print(data)
        if i == 5:
            break
    muss_pipeline.stop()





from loguru import logger
import queue
import time

import pandas as pd

from .. import generator
import threading
import importlib


def _print_extra_dep_warning(e):
    msg = (
        "Arus plugin metawear requirements are not installed. Some functionality will not be usable.\n\n"
        "Please install the metawear extra packages as follows:\n\n"
        "  pip install arus[metawear]\n\n"
    )
    print(str(e) + "\n\n" + msg)


class StartFailure(Exception):
    pass


class MetaWearScanner():
    def __init__(self):
        try:
            self._mw_discover = importlib.import_module(
                'discover', 'pymetawear')
        except ImportError as e:
            _print_extra_dep_warning(e)
            raise ImportError(e)

    def get_nearby_devices(self, max_devices=None):
        metawears = set()
        retries = 0
        max_devices = 100 if max_devices is None else max_devices
        while retries < 5 and len(metawears) < max_devices:
            logger.info('Scanning metawear devices nearby...')
            try:
                retries += 1
                candidates = self._mw_discover.discover_devices(timeout=1)
                metawears |= set(
                    map(lambda d: d[0],
                        filter(lambda d: d[1] == 'MetaWear', candidates)))
            except ValueError as e:
                logger.error(str(e))
                continue
        return list(metawears)


class MetawearTimestampCorrector(object):
    def __init__(self, sr):
        self._current_nofix_ts = None
        self._current_noloss_ts = None
        self._current_withloss_ts = None
        self._sample_interval = 1.0 / sr
        self._real_world_offset = None
        self._last_real_world_ts = None

    def _get_current_nofix_ts_in_seconds(self, data):
        return data['epoch'] / 1000.0

    def _get_real_world_offset(self, data, current_real_world_ts):
        return current_real_world_ts - \
            self._get_current_nofix_ts_in_seconds(data)

    def _apply_no_fix(self, data):
        return self._get_current_nofix_ts_in_seconds(data) + self._real_world_offset

    def _apply_fix_noloss(self, data, previous_noloss_ts):
        if previous_noloss_ts == None:
            current_noloss_ts = self._apply_no_fix(data)
        else:
            current_noloss_ts = previous_noloss_ts + self._sample_interval
        return current_noloss_ts

    def _apply_fix_withloss(self, data, previous_withloss_ts, current_real_world_ts, previous_real_world_ts):
        if previous_withloss_ts == None:
            current_withloss_ts = self._apply_fix_noloss(
                data, previous_withloss_ts)
        elif current_real_world_ts - previous_real_world_ts <= 2 * self._sample_interval:
            current_withloss_ts = self._apply_fix_noloss(
                data, previous_withloss_ts)
        elif abs(previous_withloss_ts - self._get_current_nofix_ts_in_seconds(data)) < 2 * self._sample_interval:
            # if there is no loss occuring
            current_withloss_ts = self._apply_fix_noloss(
                data, previous_withloss_ts)
        else:
            # data loss occurs
            current_nofix_ts = self._apply_no_fix(data)
            current_noloss_ts = self._apply_fix_noloss(
                data, previous_withloss_ts)
            if current_nofix_ts - current_noloss_ts > 2 * self._sample_interval:
                # late for more than two intervals, renew timestamp
                current_withloss_ts = current_nofix_ts
            else:
                current_withloss_ts = current_noloss_ts
        return current_withloss_ts

    def correct(self, data, current_real_world_ts):
        if self._real_world_offset == None:
            self._real_world_offset = self._get_real_world_offset(
                data, current_real_world_ts)
        if self._last_real_world_ts == None:
            self._last_real_world_ts = current_real_world_ts
        self._current_nofix_ts = self._apply_no_fix(data)
        self._current_noloss_ts = self._apply_fix_noloss(
            data, self._current_noloss_ts)
        self._current_withloss_ts = self._apply_fix_withloss(
            data, self._current_withloss_ts, current_real_world_ts, self._last_real_world_ts)
        diff_rw = current_real_world_ts - self._last_real_world_ts
        self._last_real_world_ts = current_real_world_ts
        return self._current_nofix_ts, self._current_noloss_ts, self._current_withloss_ts, self._get_current_nofix_ts_in_seconds(data), current_real_world_ts


class MetaWearAccelDataGenerator(generator.Generator):
    def __init__(self, addr, sr, grange, max_retries=3, **kwargs):
        super().__init__(**kwargs)

        try:
            self._mw = importlib.import_module('metawear', 'mbientlab')
            self._mw_client = importlib.import_module('client', 'pymetawear')
        except ImportError as e:
            _print_extra_dep_warning(e)
            raise ImportError(e)

        self._addr = addr
        self._sr = sr
        self._grange = grange
        self._max_retries = max_retries

        self._device = None
        self._corrector = MetawearTimestampCorrector(sr)
        self._input_count = 0
        self._internal_buffer = queue.Queue()
        self._callback_started = False
        self._start_condition = threading.Condition(threading.Lock())
        self._setup_metawear()

    def run(self, values=None, src=None, context={}):
        if self._start_metawear():
            self._generate()
        else:
            raise StartFailure(
                'Device fails to start correctly, please call generate to retry')

    def stop(self):
        self._device.led.stop_and_clear()
        time.sleep(0.5)
        self._device.accelerometer.notifications(callback=None)
        time.sleep(1)
        self._device.disconnect()
        logger.info('Disconnected.')
        self._callback_started = False
        self._internal_buffer.queue.clear()
        super().stop()

    def get_device_name(self):
        model_code = self._mw.libmetawear.mbl_mw_metawearboard_get_model(
            self._device.mw.board)
        metawear_models = self._mw.cbindings.Model()
        model_names = list(
            filter(lambda attr: '__' not in attr, dir(metawear_models)))
        for name in model_names:
            if getattr(metawear_models, name) == model_code:
                return name
        return 'NA'

    def _generate(self):
        while self._callback_started:
            if self._stop:
                break
            try:
                data = self._internal_buffer.get(timeout=0.1)
                result = self._buffering(data)
                if result is not None:
                    self._result.put((result, self._context))
            except queue.Empty:
                continue

    def _start_metawear(self):
        self._input_count = 0

        def _callback(data):
            formatted = self._format_data_as_mhealth(data)
            self._internal_buffer.put(formatted)
            if self._input_count == 0:
                logger.info('Accelerometer callback starts running...')
                with self._start_condition:
                    self._callback_started = True
                    self._start_condition.notify_all()

            self._input_count = self._input_count + 1
            if self._input_count == self._sr:
                logger.debug('Received data for one second')
                self._input_count = 1

        def _start():
            logger.info('starting accelerometer module...')
            self._device.accelerometer.notifications(
                callback=_callback)
            with self._start_condition:
                logger.info("Waiting for accelerometer callback to start...")
                self._start_condition.wait(timeout=2)
            if self._callback_started:
                logger.info(
                    'Accelerometer callback started successfully.')
                pattern = self._device.led.load_preset_pattern(
                    'solid', repeat_count=0xff)
                self._device.led.write_pattern(pattern, 'g')
                self._device.led.play()
                return True
            else:
                logger.warning(
                    'Accelerometer callback did not start successfully.')
                return False
        return _start()

    def _setup_metawear(self):
        count_retry = 0
        while self._device is None:
            try:
                self._device = self._mw_client.MetaWearClient(
                    self._addr, connect=True, debug=False)
                self._device_name = self.get_device_name()
            except Exception as e:
                if count_retry == self._max_retries:
                    raise Exception(
                        "Max retry reaches, still cannot connect to device: " + self._addr)
                logger.error(str(e))
                logger.info(str(count_retry) +
                            ' retry connect to ' + self._addr)
                count_retry = count_retry + 1
                time.sleep(1)
        logger.info("New metawear connected: {0}".format(
            self._device))
        self._mw.libmetawear.mbl_mw_metawearboard_set_time_for_response(
            self._device.mw.board, 4000)
        # high frequency throughput connection setup
        self._device.settings.set_connection_parameters(
            7.5, 7.5, 0, 6000)
        # Up to 4dB for Class 2 BLE devices
        # https://github.com/hbldh/pymetawear/blob/master/pymetawear/modules/settings.py
        # https://mbientlab.com/documents/metawear/cpp/0/settings_8h.html#a335f712d5fc0587eff9671b8b105d3ed
        # Hossain AKMM, Soh WS. A comprehensive study of Bluetooth signal parameters for localization. 2007 Ieee 18th International Symposium on Personal, Indoor and Mobile Radio Communications, Vols 1-9. 2007:428-32.
        self._device.settings.set_tx_power(power=4)

        self._device.accelerometer.set_settings(
            data_rate=self._sr, data_range=self._grange)
        self._device.accelerometer.high_frequency_stream = True

    def _format_data_as_mhealth(self, data):
        real_world_ts = time.time()
        ts_set = self._corrector.correct(data, real_world_ts)
        calibrated_values = self._calibrate_coord_system(data)
        formatted = {
            'HEADER_TIME_STAMP': [pd.Timestamp.fromtimestamp(ts_set[2])],
            'X': [calibrated_values[0]],
            'Y': [calibrated_values[1]],
            'Z': [calibrated_values[2]],
            'NO_FIX': [pd.Timestamp.fromtimestamp(ts_set[3])],
            'CLOCK': [pd.Timestamp.fromtimestamp(ts_set[4])],
            'MAC_ADDRESS': [self._addr],
            'DEVICE_NAME': [self._device_name]
        }
        return pd.DataFrame.from_dict(formatted)

    def _calibrate_coord_system(self, data):
        # axis values are calibrated according to the coordinate system of Actigraph GT9X
        # http://www.correctline.pl/wp-content/uploads/2015/01/ActiGraph_Device_Axes_Link.png
        x = data['value'].x
        y = data['value'].y
        z = data['value'].z
        if self._device_name == 'METAMOTION_R':
            # as normal wear in the case on wrist
            calibrated_x = y
            calibrated_y = -x
            calibrated_z = z
        else:
            calibrated_x = x
            calibrated_y = y
            calibrated_z = z
        return (calibrated_x, calibrated_y, calibrated_z)



import numpy as np
from scipy.signal import find_peaks
from scipy.stats import linregress
from .. import extensions as ext


EXTREMA_FEATURE_NAME_PREFIX = [
    'MAX_COUNT',
    'MIN_COUNT',
    'MAX_VAL_MEAN',
    'MAX_VAL_STD',
    'MIN_VAL_MEAN',
    'MIN_VAL_STD',
    'EXTREMA_RANGE_MEAN',
    'EXTREMA_RANGE_STD',
    'EXTREMA_RANGE_MAX',
    'EXTREMA_RANGE_MIN',
    'EXTREMA_DUR_MEAN',
    'EXTREMA_DUR_STD',
    'EXTREMA_CORR_MEAN',
    'EXTREMA_CORR_STD'
]


def extrema_features(X, sr, X_unfiltered=None, threshold=None, selected=EXTREMA_FEATURE_NAME_PREFIX):
    X = ext.numpy.atleast_float_2d(X)

    fv = []
    fv_names = []

    if EXTREMA_FEATURE_NAME_PREFIX[0] in selected:
        f, _ = extrema_count(X, threshold=threshold)
        fv += f[0, ::2].tolist()
        fv_names += [
            f'{EXTREMA_FEATURE_NAME_PREFIX[0]}_{i}' for i in range(X.shape[1])]

    if EXTREMA_FEATURE_NAME_PREFIX[1] in selected:
        f, _ = extrema_count(X, threshold=threshold)
        fv += f[0, 1::2].tolist()
        fv_names += [
            f'{EXTREMA_FEATURE_NAME_PREFIX[1]}_{i}' for i in range(X.shape[1])]

    if EXTREMA_FEATURE_NAME_PREFIX[2] in selected:
        f, _ = extrema_value(X, threshold=threshold)
        fv += f[0, ::4].tolist()
        fv_names += [
            f'{EXTREMA_FEATURE_NAME_PREFIX[2]}_{i}' for i in range(X.shape[1])]

    if EXTREMA_FEATURE_NAME_PREFIX[3] in selected:
        f, _ = extrema_value(X, threshold=threshold)
        fv += f[0, 1::4].tolist()
        fv_names += [
            f'{EXTREMA_FEATURE_NAME_PREFIX[3]}_{i}' for i in range(X.shape[1])]

    if EXTREMA_FEATURE_NAME_PREFIX[4] in selected:
        f, _ = extrema_value(X, threshold=threshold)
        fv += f[0, 2::4].tolist()
        fv_names += [
            f'{EXTREMA_FEATURE_NAME_PREFIX[4]}_{i}' for i in range(X.shape[1])]

    if EXTREMA_FEATURE_NAME_PREFIX[5] in selected:
        f, _ = extrema_value(X, threshold=threshold)
        fv += f[0, 3::4].tolist()
        fv_names += [
            f'{EXTREMA_FEATURE_NAME_PREFIX[5]}_{i}' for i in range(X.shape[1])]

    if EXTREMA_FEATURE_NAME_PREFIX[6] in selected:
        f, _ = extrema_range(X, threshold=threshold)
        fv += f[0, ::4].tolist()
        fv_names += [
            f'{EXTREMA_FEATURE_NAME_PREFIX[6]}_{i}' for i in range(X.shape[1])]

    if EXTREMA_FEATURE_NAME_PREFIX[7] in selected:
        f, _ = extrema_range(X, threshold=threshold)
        fv += f[0, 1::4].tolist()
        fv_names += [
            f'{EXTREMA_FEATURE_NAME_PREFIX[7]}_{i}' for i in range(X.shape[1])]

    if EXTREMA_FEATURE_NAME_PREFIX[8] in selected:
        f, _ = extrema_range(X, threshold=threshold)
        fv += f[0, 2::4].tolist()
        fv_names += [
            f'{EXTREMA_FEATURE_NAME_PREFIX[8]}_{i}' for i in range(X.shape[1])]

    if EXTREMA_FEATURE_NAME_PREFIX[9] in selected:
        f, _ = extrema_range(X, threshold=threshold)
        fv += f[0, 3::4].tolist()
        fv_names += [
            f'{EXTREMA_FEATURE_NAME_PREFIX[9]}_{i}' for i in range(X.shape[1])]

    if EXTREMA_FEATURE_NAME_PREFIX[10] in selected:
        f, _ = extrema_duration(X, sr=sr, threshold=threshold)
        fv += f[0, ::2].tolist()
        fv_names += [
            f'{EXTREMA_FEATURE_NAME_PREFIX[10]}_{i}' for i in range(X.shape[1])]

    if EXTREMA_FEATURE_NAME_PREFIX[11] in selected:
        f, names = extrema_duration(X, sr=sr, threshold=threshold)
        fv += f[0, 1::2].tolist()
        fv_names += [
            f'{EXTREMA_FEATURE_NAME_PREFIX[11]}_{i}' for i in range(X.shape[1])]

    if EXTREMA_FEATURE_NAME_PREFIX[12] in selected:
        f, _ = extrema_corr(
            X, sr=sr, X_unfilered=X_unfiltered, threshold=threshold)
        fv += f[0, ::2].tolist()
        fv_names += [
            f'{EXTREMA_FEATURE_NAME_PREFIX[12]}_{i}' for i in range(X.shape[1])]

    if EXTREMA_FEATURE_NAME_PREFIX[13] in selected:
        f, _ = extrema_corr(
            X, sr=sr, X_unfilered=X_unfiltered, threshold=threshold)
        fv += f[0, 1::2].tolist()
        fv_names += [
            f'{EXTREMA_FEATURE_NAME_PREFIX[13]}_{i}' for i in range(X.shape[1])]

    if len(fv) == 0:
        return None, None

    result = np.atleast_2d(fv)

    return result, fv_names


def get_extrema(x):
    maxima_inds, _ = find_peaks(x)
    minima_inds, _ = find_peaks(-x)
    return maxima_inds, minima_inds


def extrema_count(X, threshold=None):
    X = ext.numpy.atleast_float_2d(X)
    X_extrema_count = []
    names = []
    for i in range(X.shape[1]):
        max_ids, min_ids = get_extrema(X[:, i])
        X_extrema_count.append(max_ids.shape[0])
        X_extrema_count.append(min_ids.shape[0])
        names.append(f'MAX_COUNT_{i}')
        names.append(f'MIN_COUNT_{i}')
    result = ext.numpy.atleast_float_2d(X_extrema_count)
    return result, names


def extrema_value(X, threshold=None):
    X = ext.numpy.atleast_float_2d(X)
    X_extrema_value = []
    names = []
    for i in range(X.shape[1]):
        max_ids, min_ids = get_extrema(X[:, i])
        max_values, min_values = X[max_ids, i], X[min_ids, i]
        if len(max_values) == 0:
            max_value_mean = np.nan
            max_value_std = np.nan
        else:
            max_value_mean = np.nanmean(max_values)
            max_value_std = np.nanstd(max_values)
        if len(min_values) == 0:
            min_value_mean = np.nan
            min_value_std = np.nan
        else:
            min_value_mean = np.nanmean(min_values)
            min_value_std = np.nanstd(min_values)
        X_extrema_value.append(max_value_mean)
        X_extrema_value.append(max_value_std)
        X_extrema_value.append(min_value_mean)
        X_extrema_value.append(min_value_std)
        names.append(f'MAX_VAL_MEAN_{i}')
        names.append(f'MAX_VAL_STD_{i}')
        names.append(f'MIN_VAL_MEAN_{i}')
        names.append(f'MIN_VAL_STD_{i}')
    result = ext.numpy.atleast_float_2d(X_extrema_value)
    return result, names


def extrema_range(X, threshold=None):
    X = ext.numpy.atleast_float_2d(X)
    X_extrema_range = []
    names = []
    for i in range(X.shape[1]):
        max_ids, min_ids = get_extrema(X[:, i])
        extrema_ids = sorted(max_ids.tolist() + min_ids.tolist())
        extrema_ranges = np.abs(np.diff(X[extrema_ids, i]))
        if len(extrema_ranges) == 0:
            X_extrema_range += [np.nan] * 4
        else:
            X_extrema_range.append(np.nanmean(extrema_ranges))
            X_extrema_range.append(np.nanstd(extrema_ranges))
            X_extrema_range.append(np.nanmax(extrema_ranges))
            X_extrema_range.append(np.nanmin(extrema_ranges))
        names.append(f'EXTREMA_RANGE_MEAN_{i}')
        names.append(f'EXTREMA_RANGE_STD_{i}')
        names.append(f'EXTREMA_RANGE_MAX_{i}')
        names.append(f'EXTREMA_RANGE_MIN_{i}')
    result = ext.numpy.atleast_float_2d(X_extrema_range)
    return result, names


def extrema_duration(X, sr, threshold=None):
    X = ext.numpy.atleast_float_2d(X)
    X_extrema_duration = []
    names = []
    for i in range(X.shape[1]):
        max_ids, min_ids = get_extrema(X[:, i])
        extrema_ids = sorted(max_ids.tolist() + min_ids.tolist())
        extrema_durations = np.diff(extrema_ids)
        extrema_durations_mean = np.nanmean(
            extrema_durations) / float(sr)  # in seconds
        extrema_durations_std = np.nanstd(
            extrema_durations) / float(sr)  # in seconds
        X_extrema_duration.append(extrema_durations_mean)
        X_extrema_duration.append(extrema_durations_std)
        names.append(f'EXTREMA_DUR_MEAN_{i}')
        names.append(f'EXTREMA_DUR_STD_{i}')
    result = ext.numpy.atleast_float_2d(X_extrema_duration)
    return result, names


def extrema_corr(X, sr, X_unfilered=None, threshold=None):
    X = ext.numpy.atleast_float_2d(X)
    X_extrema_corr = []
    names = []
    for i in range(X.shape[1]):
        max_ids, min_ids = get_extrema(X[:, i])
        extrema_ids = sorted(max_ids.tolist() + min_ids.tolist())
        corr_list = []
        for j in range(len(extrema_ids) - 1):
            start = extrema_ids[j]
            stop = extrema_ids[j + 1]
            start_value = X[start, i]
            stop_value = X[stop, i]
            if stop - start < 2:
                continue
            x = np.arange(start=start, stop=stop)
            if X_unfilered is None:
                y_origin = X[x, i]
            else:
                y_origin = X_unfilered[x, i]
            slope, intercept, _, _, _ = linregress(
                [start, stop], [start_value, stop_value])
            y = slope * x + intercept
            corr_list.append(np.correlate(y, y_origin)[0])
        if len(corr_list) == 0:
            X_extrema_corr.append(np.nan)
            X_extrema_corr.append(np.nan)
        else:
            X_extrema_corr.append(np.nanmean(corr_list))
            X_extrema_corr.append(np.nanstd(corr_list))
        names.append(f'EXTREMA_CORR_MEAN_{i}')
        names.append(f'EXTREMA_CORR_STD_{i}')
    result = ext.numpy.atleast_float_2d(X_extrema_corr)
    return result, names



"""
Plot confusion matrix
=====================
"""

from arus.models.muss import MUSSModel
import matplotlib.pyplot as plt
muss = MUSSModel()
input_class = ['Sit', 'Stand', 'Stand', 'Walk']
predict_class = ['Sit', 'Walk', 'Stand', 'Stand']
fig = muss.get_confusion_matrix(input_class, predict_class, labels=[
    'Sit', 'Stand', 'Walk'], graph=True)

df = muss.get_confusion_matrix(input_class, predict_class, labels=[
    'Sit', 'Stand', 'Walk'], graph=False)

print(df)



import pandas as pd
import pytest

from .. import dataset as ds


@pytest.fixture(scope="module")
def spades_lab_data():
    return ds.load_dataset('spades_lab')


@pytest.fixture(scope="module")
def spades_lab_ds():
    return ds.MHDataset(path=ds.get_dataset_path('spades_lab'),
                        name='spades_lab', input_type=ds.InputType.MHEALTH_FORMAT)


@pytest.fixture(scope="module")
def single_mhealth_sensor(spades_lab_ds):
    sensor = spades_lab_ds.get_sensors(pid='SPADES_1', placement='DW')[0]
    sensor_file = sensor.paths[0]
    data = pd.read_csv(sensor_file, parse_dates=[0])
    return data



from .. import mhealth_format as mh
from .. import dataset as ds
from dataclasses import dataclass, field
import typing
from loguru import logger
from .. import error_code

from sklearn.metrics import confusion_matrix
import pandas as pd


@dataclass
class HARModel:
    mid: str
    name: str = "HAR"
    model: typing.Any = None
    train_perf: typing.Dict[str, typing.Any] = field(default_factory=dict)
    train_pids: typing.List[str] = field(default_factory=list)
    data_set: ds.MHDataset = field(init=False)

    def __post_init__(self):
        logger.info(self)

    def reset_model(self):
        self.model = None

    def load_dataset(self, data: ds.MHDataset):
        self.data_set = data

    def compute_features(self, **kwargs):
        raise NotImplementedError(
            'This method must be implemented by sub classes!')

    def compute_class_set(self, **kwargs):
        raise NotImplementedError(
            'This method must be implemented by sub classes!')

    def train(self, **kwargs):
        raise NotImplementedError(
            'This method must be implemented by sub classes!')

    def predict(self, *input_objs, **kwargs):
        raise NotImplementedError(
            'This method must be implemented by sub classes!')

    def cross_validation(self, pids=None, n_fold=10, **kwargs):
        raise NotImplementedError(
            'This method must be implemented by sub classes!')

    def logo_validation(self, pids=None, group_col=None, **kwargs):
        raise NotImplementedError(
            'This method must be implemented by sub classes!')

    def learning_curve_logo(self, pids=None, group_col=None, **kwargs):
        raise NotImplementedError(
            'This method must be implemented by sub classes')

    def learning_curve_cv(self, pids=None, n_fold=10, **kwargs):
        raise NotImplementedError(
            'This method must be implemented by sub classes')

    @staticmethod
    def confusion_matrix(true_labels, pred_labels, label_names):
        cm = confusion_matrix(true_labels, pred_labels, labels=label_names)
        cm_df = pd.DataFrame(data=cm, index=label_names, columns=label_names)
        return cm_df

    @staticmethod
    def ignore_classes(fcs, task_name, remove_classes=['Unknown', 'Transition']):
        is_valid_label = ~fcs[task_name].isin(remove_classes).values
        fcs = fcs.loc[is_valid_label, :]
        return fcs



from .. import moment
from .. import mhealth_format as mh
import numpy as np
import pkg_resources
import pandas as pd
from loguru import logger


def _get_annotation_durations(annot_df):
    durations = annot_df.groupby(annot_df.columns[3]).apply(
        lambda rows: np.sum(rows.iloc[:, 2] - rows.iloc[:, 1]))
    return durations


def _annot_to_activity(annot_df, pid, start_time, stop_time):
    if annot_df is None:
        return "Unknown"
    if annot_df.shape[0] == 0:
        return "Unknown"
    label_list = annot_df[mh.ANNOTATION_LABEL_COL].str.lower()
    annot_df[mh.ANNOTATION_LABEL_COL] = label_list
    annot_df = annot_df.loc[(label_list != 'wear on')
                            & (label_list != 'wearon'), :]
    if annot_df.shape[0] == 0:
        return "Unknown"
    labels = annot_df.iloc[:, 3].unique()
    labels.sort()
    label = ' '.join(labels).lower().strip()

    # filter if it does not cover the entire 12.8s
    durations = _get_annotation_durations(annot_df)
    interval = int(moment.Moment.get_duration(
        start_time, stop_time, unit='ms'))

    if not np.all(durations.values >= np.timedelta64(interval, 'ms')):
        return "Transition"
    # special cases
    if pid == 'SPADES_26':
        if 'biking' in label and start_time.hour == 11 and stop_time.minute > 26:
            return "Stationary cycle ergometry"
    elif pid == 'SPADES_19':
        if '3 mph' in label and 'arms on desk' in label and 'treadmill' in label:
            return "Level treadmill walking at 3 mph with arms on desk"

    if 'stairs' in label and 'up' in label:
        return 'Walking upstairs'
    elif 'stairs' in label and 'down' in label:
        return 'Walking downstairs'
    if 'mbta' in label or 'city' in label or 'outdoor' in label:
        return 'Unknown'
    if "sitting" in label and 'writing' in label:
        return 'Sitting and writing'
    elif 'stand' in label and 'writ' in label:
        return 'Standing and writing at a table'
    elif 'sit' in label and ('web' in label or 'typ' in label):
        return 'Sitting and typing on a keyboard'
    elif 'reclin' in label and ('text' in label or 'web' in label):
        return 'Reclining and using phone'
    elif 'sit' in label and 'story' in label and ('city' not in label and 'outdoor' not in label):
        return "Sitting and talking"
    elif "reclin" in label and 'story' in label:
        return 'Reclining and talking'
    elif "stand" in label and ('web' in label or 'typ' in label):
        return "Standing and typing on a keyboard"
    elif 'bik' in label and ('stationary' in label or '300' in label):
        return "Stationary cycle ergometry"
    elif ('treadmill' in label or 'walk' in label) and '1' in label:
        return "Level treadmill walking at 1 mph with arms on desk"
    elif ('treadmill' in label or 'walk' in label) and '2' in label:
        return "Level treadmill walking at 2 mph with arms on desk"
    elif 'treadmill' in label and 'phone' in label:
        return "Level treadmill walking at 3-3.5 mph while holding a phone to the ear and talking"
    elif 'treadmill' in label and 'bag' in label:
        return "Level treadmill walking at 3-3.5 mph and carrying a bag"
    elif 'treadmill' in label and 'story' in label:
        return "Level treadmill walking at 3-3.5 mph while talking"
    elif ('treadmill' in label or 'walk' in label) and 'drink' in label:
        return 'Level treadmill walking at 3-3.5 mph and carrying a drink'
    elif ('treadmill' in label or 'walk' in label) and ('3.5' in label or '3' in label):
        return 'Level treadmill walking at 3-3.5 mph'
    elif '5.5' in label or 'jog' in label or 'run' in label:
        return 'Treadmill running at 5.5 mph & 5% grade'
    elif 'laundry' in label:
        return 'Standing and folding towels'
    elif 'sweep' in label:
        return 'Standing and sweeping'
    elif 'shelf' in label and 'load' in label:
        return 'Standing loading/unloading shelf'
    elif 'lying' in label:
        return "Lying on the back"
    elif label == 'sitting' or ('sit' in label and 'still' in label):
        return "Sitting still"
    elif label == "still" or 'standing' == label or label == 'standing still':
        return "Self-selected free standing"
    else:
        return 'Unknown'


def _load_task_class_map():
    map_filepath = pkg_resources.resource_filename(
        'arus', 'spades_lab/task_class_map.csv')
    return pd.read_csv(map_filepath)


def class_set(*annot_dfs, task_names=None, st=None, et=None, pid=None, **kwargs):
    class_labels = {mh.TIMESTAMP_COL: [st],
                    mh.START_TIME_COL: [st],
                    mh.STOP_TIME_COL: [et]}

    task_class_map = _load_task_class_map()
    annot_df = annot_dfs[0]
    activity = _annot_to_activity(annot_df, pid, st, et)

    for task_name in task_names:
        if task_name not in task_class_map.columns:
            logger.warning(
                f"{task_name} is not a valid task name for the current dataset")
        else:
            class_label = task_class_map.loc[task_class_map.ACTIVITY ==
                                             activity, task_name]
            class_labels[task_name] = class_label.values.tolist()
    class_vector = pd.DataFrame.from_dict(class_labels)
    return class_vector



"""
Demonstration of the usage of arus.core.Pipeline (1)
====================================================

The pipeline uses one annotation stream and one sensor stream as input.
"""

from arus.core.pipeline import Pipeline
import arus
import datetime as dt
import pandas as pd
import numpy as np


def _pipeline_test_processor(chunk_list, **kwargs):
    import pandas as pd
    import numpy as np
    result = {'NAME': [],
              'START_TIME': [], 'STOP_TIME': [], 'VALUE': []}
    for data, st, et, prev_st, prev_et, name in chunk_list:
        if name == 'sensor-stream':
            result['NAME'].append(name)
            result['START_TIME'].append(data.iloc[0, 0])
            result['STOP_TIME'].append(data.iloc[-1, 0])
            result['VALUE'].append(np.mean(data.iloc[:, 1:].values))
        elif name == 'annotation-stream':
            result['NAME'].append(name)
            result['START_TIME'].append(data.iloc[0, 1])
            result['STOP_TIME'].append(data.iloc[-1, 2])
            result['VALUE'].append('-'.join(data.iloc[:, 3].values))
    result = pd.DataFrame.from_dict(result)
    return result


if __name__ == "__main__":
    window_size = 12.8
    sr = 80

    gr1 = arus.generator.RandomAccelDataGenerator(
        sr=sr, grange=8, sigma=1, buffer_size=100)
    seg1 = arus.segmentor.SlidingWindowSegmentor(window_size)
    stream1 = arus.Stream(gr1, seg1, name='sensor-stream')

    gr2 = arus.generator.RandomAnnotationDataGenerator(
        labels=["Sitting", 'Standing', 'Lying', 'Walking', 'Running'], duration_mu=8, duration_sigma=2, num_mu=3, buffer_size=100)
    seg2 = arus.segmentor.SlidingWindowSegmentor(
        window_size, st_col=1, et_col=2)
    stream2 = arus.Stream(gr2, seg2, name='annotation-stream')

    start_time = dt.datetime.now()
    pipeline = Pipeline(max_processes=2)
    pipeline.add_stream(stream1)
    pipeline.add_stream(stream2)
    pipeline.set_processor(_pipeline_test_processor)
    pipeline.start(start_time=start_time)
    results = []
    for result, st, et, prev_st, prev_et, name in pipeline.get_iterator():
        print(st)
        result['WINDOW_ST'] = st
        result['WINDOW_ET'] = et
        result['PREV_WINDOW_ST'] = prev_st
        result['PREV_WINDOW_ET'] = prev_et
        result['STREAM_NAME'] = name
        results.append(result)
        if len(results) == 10:
            break
    pipeline.stop()
    print(pd.concat(results, axis=0, sort=False))



from .. import mhealth_format as mh
from .. import extensions as ext
from .. import moment
import pandas as pd


def preprocess_processor(chunk_list, **kwargs):
    import pandas as pd
    from .. import dataset
    from .nn import preprocess
    results = []
    for df, st, et, prev_st, prev_et, name in chunk_list:
        if name in mh.SENSOR_PLACEMENTS:  # process sensors
            if df.empty:
                continue
            else:
                result = preprocess(
                    df, name=name, sr=kwargs[name]['sr'], st=st, et=et)
                if result is not None:
                    results.append(result)
    if len(results) == 0:
        combined_df = pd.DataFrame()
    elif len(results) < len(chunk_list):
        combined_df = pd.DataFrame()
    else:
        combined_df = pd.concat(results, axis=1, sort=False, )
    return combined_df


def preprocess(df, name, sr, st, et):
    beginning_delay = moment.Moment.get_duration(st, df.iloc[0, 0], unit='s')
    end_delay = moment.Moment.get_duration(df.iloc[-1, 0], et, unit='s')
    if beginning_delay > 1 or end_delay > 1:
        return None
    col_names = ['TS', name + '_0', name + '_1', name + '_2']
    ts = moment.Moment.seq_to_unix_timestamp(df['HEADER_TIME_STAMP'].values)
    X = df.iloc[:, 1:].values
    st = moment.Moment(st).to_unix_timestamp()
    et = moment.Moment(et).to_unix_timestamp()
    new_ts, new_X = ext.numpy.regularize_sr(ts, X, sr, st=st, et=et)
    new_df = pd.DataFrame(index=new_ts, data=new_X).reset_index(drop=False)
    new_df.columns = col_names
    new_df = new_df.set_index('TS')
    return new_df



import PySimpleGUI as sg
import time
import numpy as np

from matplotlib import pyplot as plt
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg

HEADING_FONT = ('Helvetica', 12, 'bold')
PRIMARY_FONT = ('Helvetica', 10)
SMALL_FONT = ('Helvetica', 8)
BIG_FONT = ('Helvetica', 14, 'bold')


def heading(text, fixed_column_width=None, key=None):
    return sg.Text(text=text,
                   relief=sg.RELIEF_FLAT,
                   font=HEADING_FONT,
                   auto_size_text=True,
                   justification='center',
                   size=(fixed_column_width, None),
                   key=key)


def text(text, fixed_column_width=None, text_color=None, background_color=None, key=None):
    return sg.Text(text=text,
                   font=PRIMARY_FONT,
                   size=(fixed_column_width, None),
                   text_color=text_color,
                   background_color=background_color,
                   key=key)


def big_text(text, fixed_column_width=None, text_color=None, background_color=None, key=None):
    return sg.Text(text=text,
                   font=HEADING_FONT,
                   size=(fixed_column_width, None),
                   text_color=text_color,
                   background_color=background_color,
                   key=key)


def selection_list(items, default_selections=None, mode='single', fixed_column_width=None, rows=20, key=None):
    if mode == 'single':
        mode = sg.LISTBOX_SELECT_MODE_SINGLE
    else:
        mode = sg.LISTBOX_SELECT_MODE_EXTENDED
    return sg.Listbox(values=items,
                      default_values=default_selections,
                      select_mode=mode,
                      font=PRIMARY_FONT,
                      size=(fixed_column_width, rows),
                      key=key,
                      enable_events=False)


def dropdown_list(items, default_item=None, fixed_column_width=None, key=None):
    return sg.Combo(items, default_value=default_item, size=(fixed_column_width, None), key=key, enable_events=True)


def control_button(text, size='normal', fixed_column_width=None, disabled=False, key=None):
    if size == 'normal':
        font = PRIMARY_FONT
    elif size == 'big':
        font = HEADING_FONT
    elif size == 'small':
        font = SMALL_FONT
    return sg.Button(button_text=text,
                     font=font,
                     auto_size_button=True,
                     size=(fixed_column_width, None),
                     key=key,
                     disabled=disabled)


class TextInput:
    def __init__(self, confirm_button_text, candidate_items=[], fixed_column_width=None, confirm_button_disabled=False, text_input_disabled=False, confirm_button_key=None, text_input_key=None):
        if fixed_column_width is not None:
            button_width = int(fixed_column_width * 1 / 4)
            input_width = int(fixed_column_width * 3 / 4)
        else:
            button_width = None
            input_width = None
        self._confirm_button = control_button(
            confirm_button_text,
            size='normal',
            fixed_column_width=button_width,
            disabled=confirm_button_disabled,
            key=confirm_button_key)
        self._text_input = sg.Combo(
            values=candidate_items,
            font=PRIMARY_FONT,
            size=(input_width, None),
            key=text_input_key,
            disabled=text_input_disabled
        )

    def get_component_as_row(self):
        return [self._text_input, self._confirm_button]

    def update_button(self, **kwargs):
        self._confirm_button.update(**kwargs)

    def update_input(self, **kwargs):
        self._text_input.update(**kwargs)


class ProgressBar:
    def __init__(self, text, fixed_column_width=None, text_key=None, bar_key=None):
        self._progress_text = sg.Text(text,
                                      key=text_key,
                                      font=PRIMARY_FONT,
                                      size=(fixed_column_width, None))

        self._progress_bar = sg.ProgressBar(100,
                                            orientation='h',
                                            auto_size_text=True,
                                            size=(fixed_column_width, 20),
                                            key=bar_key)

    def get_component(self):
        return [self._progress_text], [self._progress_bar]

    def update(self, percentage, text=None):
        if text is not None:
            self._progress_text.update(value=text)
        self._progress_bar.update_bar(int(100 * percentage))

    def increment(self, text=None):
        if text is not None:
            self._progress_text.update(value=text)
        self._progress_bar.update_bar(
            self._progress_bar.TKProgressBar.TKProgressBarForReal['value'] + 1)


def checkbox(text, fixed_column_width=None, default_checked=False, disabled=False, key=None):
    return sg.Checkbox(text,
                       default=default_checked,
                       disabled=disabled,
                       auto_size_text=True,
                       size=(fixed_column_width, None),
                       font=PRIMARY_FONT,
                       key=key, enable_events=True)


class RadioGroup:
    def __init__(self, group_id, n, default_index=0, direction='horizontal',
                 fixed_column_width=None):
        self._group_id = group_id
        self._direction = direction
        self._radio_boxes = []
        self._els = []
        self._n = n
        self._default_index = default_index
        if fixed_column_width is not None:
            if self._direction == 'horizontal':
                self._width = int(
                    fixed_column_width / n)
            else:
                self._width = fixed_column_width
        else:
            self._width = None

    def add_radiobox(self, text, key, disabled=False):
        if len(self._radio_boxes) == self._default_index:
            checked = True
        else:
            checked = False
        box = sg.Radio(text, self._group_id, default=checked, disabled=disabled,
                       auto_size_text=True, font=PRIMARY_FONT, size=(self._width, None), enable_events=False, key=key)
        self._els.append(box)
        if self._direction == 'horizontal':
            self._radio_boxes.append(box)
        else:
            self._radio_boxes = self._radio_boxes + [box]

    def add_radioboxes(self, texts, keys, disable_states):
        for text, key, disabled in zip(texts, keys, disable_states):
            self.add_radiobox(text, key, disabled=disabled)

    def disable_all(self):
        for el in self._els:
            el.update(disabled=True)

    def enable_all(self):
        for el in self._els:
            el.update(disabled=False)

    def get_component(self):
        return self._radio_boxes


def table(cells, headings, fixed_column_width=None, key=None):
    return sg.Table(values=cells,
                    headings=headings,
                    size=(fixed_column_width, None),
                    def_col_width=int(fixed_column_width/len(headings)),
                    auto_size_columns=False,
                    key=key)


def image(img_url, key=None):
    return sg.Image(img_url,
                    enable_events=False,
                    key=key)


class Plot:
    def __init__(self, fixed_column_width=None, rows=None, key=None):
        self._el = sg.Canvas(size=(fixed_column_width, rows), key=None)
        self._graph = None
        self._fig = None

    def update_plot(self, fig=None):
        if fig is not None:
            self._fig = fig
            _, _, figure_w, figure_h = fig.bbox.bounds
            self._el.set_size((figure_w, figure_h))
            self._graph = FigureCanvasTkAgg(fig, self._el.TKCanvas)
            self._graph.draw()
            self._graph.get_tk_widget().pack(side='top', fill='both', expand=1)
        else:
            self._graph.draw()
            self._graph.get_tk_widget().pack(side='top', fill='both', expand=1)

    def get_figure(self):
        return self._fig

    def get_component(self):
        return self._el


class DeviceInfo:
    def __init__(self, device_name, device_addr_list=[], placement_img_url=None, fixed_column_width=None, device_name_key=None, placement_img_key=None, device_addr_key=None, device_selected=True, device_selection_disabled=False):
        self._device_name = checkbox(device_name,
                                     fixed_column_width=fixed_column_width, default_checked=device_selected, key=device_name_key, disabled=device_selection_disabled)
        self._device_placement = image(placement_img_url,
                                       key=placement_img_key)
        self._device_addr_list = dropdown_list(
            items=device_addr_list, default_item=None, fixed_column_width=fixed_column_width, key=device_addr_key)

    def get_component(self):
        return (
            [self._device_name],
            [self._device_placement],
            [self._device_addr_list]
        )

    def update_addr_list(self, addrs, index):
        self._device_addr_list.update(values=addrs, set_to_index=index)

    def update_addr_selection(self, index):
        self._device_addr_list.update(set_to_index=index)

    def enable_selection(self):
        self._device_name.update(disabled=False)

    def disable_selection(self):
        self._device_name.update(disabled=True)

    def disable_addr_list(self):
        self._device_addr_list.update(disabled=True)

    def enable_addr_list(self):
        self._device_addr_list.update(disabled=False)

    def update_selection(self, selected=False):
        self._device_name.update(selected)


class LabelGrid:
    def __init__(self, n, fixed_column_width=None, default_background_color=None, default_text_color=None, n_cols=4):
        self._n = n
        if fixed_column_width is not None:
            self._width = int(fixed_column_width / n_cols)
        else:
            self._width = None
        self._default_background_color = default_background_color
        self._default_text_color = default_text_color
        self._n_cols = n_cols
        self._label_grid = [[]]

    def add_label(self, label, key=None, background_color=None, text_color=None):
        el = big_text(label,
                      background_color=background_color or self._default_background_color,
                      text_color=text_color or self._default_text_color, fixed_column_width=self._width,
                      key=key)
        last_row = self._label_grid[-1]
        if len(last_row) < self._n_cols:
            self._label_grid[-1].append(el)
        else:
            self._label_grid.append([el])

    def add_labels(self, labels, keys, background_colors, text_colors):
        for label, key, bcolor, tcolor in zip(labels, keys, background_colors, text_colors):
            self.add_label(label, key, background_color=bcolor,
                           text_color=tcolor)

    def get_component(self):
        return self._label_grid

    def update_label_by_key(self, key, **kwargs):
        for row in self._label_grid:
            for el in row:
                if el.Key == key:
                    el.update(**kwargs)
                    break
            break

    def update_label_by_index(self, i, j, **kwargs):
        self._label_grid[i][j].update(**kwargs)

    def update_labels(self, new_labels, new_b_colors, new_t_colors):
        i = 0
        for row in self._label_grid:
            for el in row:
                el.update(
                    new_labels[i], background_color=new_b_colors[i], text_color=new_t_colors[i])
                i = i + 1


class Timer:
    def __init__(self):
        self.reset()

    def is_running(self):
        return len(self._start_time) > 0 and len(self._start_time) > len(self._stop_time)

    def start(self):
        self._start_time.append(time.time())
        self._last_tick = self._start_time[0]

    def get_total_lapsed_time(self, formatted=False):
        if len(self._start_time) - len(self._stop_time) == 1:
            stop_time = self._stop_time + [time.time()]
        else:
            stop_time = self._stop_time
        total_time = 0
        for st, et in zip(self._start_time, stop_time):
            total_time += (et - st)
        if formatted:
            total_time = self.format_lapse_time(total_time)
        return total_time

    def get_lapsed_times(self, formatted=False):
        lapse_times = []
        if len(self._start_time) - len(self._stop_time) == 1:
            stop_time = self._stop_time + [time.time()]
        else:
            stop_time = self._stop_time
        for st, et in zip(self._start_time, stop_time):
            if formatted:
                ts = self.format_lapse_time(et - st)
            else:
                ts = et - st
            lapse_times.append(ts)
        return lapse_times

    def get_last_lapse_time(self, formatted=False):
        if len(self._start_time) - len(self._stop_time) == 1:
            stop_time = self._stop_time + [time.time()]
        else:
            stop_time = self._stop_time
        ts = stop_time[-1] - self._start_time[-1]
        if formatted:
            return self.format_lapse_time(ts)
        else:
            return ts

    def format_lapse_time(self, ts):
        seconds = np.floor(ts)
        sec_digits = str(int(seconds % 60))
        min_digits = str(int(np.floor(seconds / 60) % 60))
        hour_digits = str(int(np.floor(seconds / 3600) % 60))

        if len(sec_digits) == 1:
            sec_digits = '0' + sec_digits
        if len(min_digits) == 1:
            min_digits = '0' + min_digits
        if len(hour_digits) == 1:
            hour_digits = '0' + hour_digits
        return hour_digits + ':' + min_digits + ':' + sec_digits

    def update(self):
        self._marker_time.append(time.time())

    def tick(self):
        self._last_tick = time.time()

    def stop(self):
        if len(self._stop_time) == len(self._start_time):
            return
        self._stop_time.append(time.time())

    def reset(self):
        self._start_time = []
        self._stop_time = []
        self._marker_time = []
        self._last_tick = None

    def since_last_tick(self):
        return time.time() - self._last_tick



import queue
import abc


class Operator(abc.ABC):
    def __init__(self):
        self._context = {}
        self._stop = False
        self._result = queue.Queue()

    @abc.abstractmethod
    def run(self, *, values=None, src=None, context={}):
        pass

    def set_context(self, **context):
        self._context = context

    def merge_context(self, added_context):
        new_context = {**self._context, **added_context}
        return new_context

    def stop(self):
        self._stop = True

    def get_result(self):
        try:
            while True:
                if self._stop:
                    break
                result, new_context = self._result.get(timeout=0.1)
                yield result, new_context
        except queue.Empty:
            pass



import queue
import threading
import time
import enum
from loguru import logger
import typing

from . import node
from . import operator


class Stream(operator.Operator):
    """
    The base class for data stream.

    Stream class is an abstraction of any data source that can be loaded into memory in arbitrary chunk size either asynchronously (currently only support threading) or synchronously.
    """

    def __init__(self, generator: "arus.generator.Generator", segmentor: "arus.segmentor.Segmentor", name: typing.Optional[str] = 'default-stream'):
        """

        Arguments:
            generator: a Generator instance that can provide streaming data.
            segmentor: a Segmentor instance that is responsible for segmenting the streaming data into chunks.
            name: The name of the data stream will also be used as the name of the sub-thread that is used to load data. Defaults to 'default-stream'.
        """
        super().__init__()
        self._name = name
        self._generator = node.Node(op=generator, t=node.Node.Type.INPUT,
                                    name=self._name + '-generator')
        self._segmentor = node.Node(op=segmentor, t=node.Node.Type.PIPE,
                                    name=self._name + '-segmentor')
        self.set_essential_context(stream_id=name)

    def run(self, *, values=None, src=None, context={}):
        logger.info('Stream is starting.')
        self._segmentor.get_op().set_ref_time(self._context['ref_start_time'])
        self._generator.get_op().set_context(data_id=self._context['data_id'])
        self._segmentor.start()
        self._generator.start()
        logger.info('Stream started.')

    def set_essential_context(self, start_time=None, stream_id=None):
        self._context['ref_start_time'] = start_time
        self._context['data_id'] = stream_id

    def start(self, start_time: "str, datetime, numpy.datetime64, pandas.Timestamp" = None):
        """Method to start loading data from the provided data source.

        Arguments:
            start_time: The reference time for segmentation.

        Note:
            `start_time` is used to sync between multiple streams. If it is `None`, the default value would be extracted from the first sample of the loaded data.
        """
        self._context['ref_start_time'] = start_time
        self.run()

    def stop(self):
        """Stop the loading process."""
        logger.info('Stream is stopping.')
        self._segmentor.stop()
        logger.info('Segmentor thread stopped.')
        time.sleep(0.1)
        self._generator.stop()
        logger.info('Generator thread stopped.')
        super().stop()
        logger.info('Stream stopped.')

    def get_result(self):
        while True:
            if self._stop:
                break
            data = next(self._generator.produce())
            self._segmentor.consume(data)
            data = next(self._segmentor.produce())
            if data.signal == node.Node.Signal.WAIT:
                pass
            elif data.signal == node.Node.Signal.DATA:
                yield data.values, data.context
            else:
                pass




DOMINANT_WRIST = 'DW'
NON_DOMINANT_WRIST = 'NDW'
DOMINANT_ANKLE = 'DA'
NON_DOMINANT_ANKLE = 'NDA'
DOMINANT_THIGH = 'DT'
NON_DOMINANT_THIGH = 'NDT'
DOMINANT_HIP = 'DH'
NON_DOMINANT_HIP = 'NDH'

TIMESTAMP_COL = 'HEADER_TIME_STAMP'
START_TIME_COL = 'START_TIME'
STOP_TIME_COL = 'STOP_TIME'
FEATURE_SET_TIMESTAMP_COLS = [TIMESTAMP_COL, START_TIME_COL, STOP_TIME_COL]
FEATURE_SET_PID_COL = 'PID'
FEATURE_SET_PLACEMENT_COL = 'PLACEMENT'
ANNOTATION_LABEL_COL = 'LABEL_NAME'

META_FOLDER = 'MetaCrossParticipants'
PROCESSED_FOLDER = 'DerivedCrossParticipants'
MASTER_FOLDER = "MasterSynced"
DERIVED_FOLDER = 'Derived'
SUBJECT_META_FOLDER = 'Meta'
SUBJECT_LOG_FOLDER = 'Logs'

META_LOCATION_MAPPING_FILENAME = 'location_mapping.csv'
META_SUBJECTS_FILENAME = 'subjects.csv'
META_CLASS_CATEGORY = 'class_category.csv'

FILE_TIMESTAMP_FORMAT = '%Y-%m-%d-%H-%M-%S-%f'

SENSOR_FILE_TYPE = 'sensor'
ANNOTATION_FILE_TYPE = 'annotation'
FEATURE_SET_FILE_TYPE = 'set'
FEATURE_FILE_TYPE = 'feature'
CLASS_FILE_TYPE = 'class'

FILE_TYPES = [SENSOR_FILE_TYPE, ANNOTATION_FILE_TYPE,
              FEATURE_SET_FILE_TYPE, FEATURE_FILE_TYPE, CLASS_FILE_TYPE]

MHEALTH_FLAT_FILEPATH_PATTERN = r'(\w+)[\/\\]{1}(?:(?:MasterSynced[\/\\]{1})|(?:Derived[\/\\]{1}(?:\w+[\/\\]{1})*))[0-9A-Za-z\-\.]+\.csv(\.gz)*'
MHEALTH_FILEPATH_PATTERN = r'(\w+)[\/\\]{1}(?:(?:MasterSynced[\/\\]{1})|(?:Derived[\/\\]{1}(?:\w+[\/\\]{1})*))\d{4}[\/\\]{1}\d{2}[\/\\]{1}\d{2}[\/\\]{1}\d{2}'
CAMELCASE_PATTERN = r'(?:[A-Z][A-Za-z0-9]+)+'
VERSIONCODE_PATTERN = r'(?:NA|[0-9x]+)'
SID_PATTERN = r'[A-Z0-9]+'
ANNOTATOR_PATTERN = r'[A-Za-z0-9]+'
FILE_TIMESTAMP_PATTERN = r'[0-9]{4}(?:\-[0-9]{2}){5}-[0-9]{3}-(?:P|M)[0-9]{4}'
FILE_EXTENSION_PATTERN = r'''
(?:sensor|event|log|annotation|feature|class|prediction|model|classmap)\.csv
'''



import pandas as pd
import os
from loguru import logger
import tqdm
import datetime
import numpy as np
from . import actigraph
from .. import mhealth_format as mh
from ..extensions.pandas import segment_by_time
import math
import enum


class FileType(enum.Enum):
    SENSOR = enum.auto(),
    ANNOTATION = enum.auto()


def format_time(ts):
    s = ts.strftime('%Y-%m-%d %H:%M:%S')
    ms = '{:03.0f}'.format(math.floor(ts.microsecond / 1000.0))
    result = "{}.{}".format(s, ms)
    return result


def signify_annotation_dataframes(*dfs, data_id, output_path, session_span=None):
    # merge and sort
    merged = pd.concat(dfs, axis=0)
    merged = merged.sort_values(by=['START_TIME'])
    merged = merged.loc[merged['START_TIME']
                        != merged['STOP_TIME'], :]
    # segment by session span
    if session_span is not None:
        segmented = segment_by_time(
            merged, seg_st=session_span[0], seg_et=session_span[1], st_col=0, et_col=1)
    else:
        segmented = merged

    # save as signaligner label file
    save_as_signaligner(segmented, output_path,
                        FileType.ANNOTATION, labelset=data_id, mode='w', index=False, header=True)


def signify_annotation_files(filepaths, data_id, output_path, session_span):
    dfs = []
    with tqdm.tqdm(total=len(filepaths)) as bar:
        for filepath in filepaths:
            bar.set_description(
                "Convert file to signaligner: {}".format(filepath))
            df = pd.read_csv(
                filepath, header=0, sep=',', compression="infer", quotechar='"', parse_dates=[0, 1, 2], infer_datetime_format=True)
            dfs.append(df)
            bar.update()
    signify_annotation_dataframes(
        *dfs, data_id=data_id, output_path=output_path, session_span=session_span)


def auto_split_session_span(session_span, auto_range='W-SUN'):
    sub_sessions = pd.date_range(session_span[0], session_span[1],
                                 freq=auto_range, normalize=True).to_pydatetime().tolist()
    if len(sub_sessions) == 0:
        sub_sessions = list(session_span)
    else:
        if sub_sessions[0].strftime('%Y-%m-%d-%H') != session_span[0].strftime('%Y-%m-%d-%H'):
            sub_sessions = [session_span[0]] + sub_sessions
        if sub_sessions[-1].strftime('%Y-%m-%d-%H') != session_span[1].strftime('%Y-%m-%d-%H'):
            sub_sessions += [session_span[1]]
    return sub_sessions


def shrink_session_span(session_span, date_range=None):
    if date_range is None:
        return session_span
    else:
        if len(date_range) == 1:
            st = datetime.datetime.strptime(date_range[0], '%Y-%m-%d')
            et = session_span[1]
        elif len(date_range) == 2:
            if date_range[0] == '':
                et = datetime.datetime.strptime(date_range[1], '%Y-%m-%d')
                st = session_span[0]
            else:
                st = datetime.datetime.strptime(date_range[0], '%Y-%m-%d')
                et = datetime.datetime.strptime(date_range[1], '%Y-%m-%d')
        if st > session_span[1] or et < session_span[0]:
            logger.warning(
                'Input date range is beyond the available date range of the dataset, ignore it')
            return session_span
        st = max(session_span[0], st or session_span[0])
        et = min(session_span[1], et or session_span[1])
        return st, et


def signify_sensor_files(filepaths, data_id, output_path, output_annotation_path, session_span, sr):
    hourly_markers = pd.date_range(session_span[0], session_span[1],
                                   freq='1H', closed='left').to_pydatetime().tolist()
    n = len(hourly_markers)
    if os.path.exists(output_path):
        logger.info('Remove the existing file')
        os.remove(output_path)
    if os.path.exists(output_annotation_path):
        logger.info(
            'Remove the existing missing data annotation file')
        os.remove(output_annotation_path)
    last_row = None
    with tqdm.tqdm(total=n) as bar:
        for marker in hourly_markers:
            bar.set_description(
                "Convert sensor data to signaligner format: {}".format(marker))
            date_str = "{}-{:02d}-{:02d}-{:02d}".format(marker.year,
                                                        marker.month, marker.day, marker.hour)
            selected_files = list(filter(lambda f: date_str in f, filepaths))
            if len(selected_files) == 0:
                filepath = None
            elif len(selected_files) == 1:
                filepath = selected_files[0]
            else:
                logger.warning(
                    'Multiple files found: {}'.format(selected_files))
                filepath = selected_files[0]
            hourly_df = _regularize_samples(marker, filepath, sr, data_id)

            if hourly_df.iloc[0, :].isna().any() and last_row is not None:
                hourly_df.iloc[0, :] = last_row.values
                last_row = None
            if hourly_df.iloc[-1, :].notna().all():
                last_row = hourly_df.iloc[-1, :]
            hourly_df = hourly_df.iloc[:-1, :]
            save_as_signaligner(hourly_df,
                                output_path,
                                file_type=FileType.SENSOR,
                                sid=data_id.split('-')[0],
                                session_st=session_span[0], session_et=session_span[1],
                                sr=sr,
                                mode='a',
                                index=False,
                                header=False,
                                float_format='%.6f')

            annotation_df = _data_to_annotation(hourly_df, sr=sr)
            if not os.path.exists(output_annotation_path):
                save_as_signaligner(annotation_df, output_annotation_path,
                                    FileType.ANNOTATION, labelset=data_id, mode='w', header=True, index=False)
            else:
                save_as_signaligner(annotation_df, output_annotation_path,
                                    FileType.ANNOTATION, labelset=data_id, mode='a', header=False, index=False)
            bar.update()


def _regularize_samples(start_time, filepath=None, sr=50, data_id=None):
    freq = str(1000 / sr) + 'L'
    tolerance = str(500 / sr) + 'L'
    sample_ts = pd.date_range(start_time, start_time +
                              datetime.timedelta(hours=1), freq=freq)

    if data_id is not None:
        data_type = data_id.split('-')[1]
        col_names = mh.parse_column_names_from_data_type(data_type)
    if filepath is None:
        out_df = sample_ts.to_frame(index=False)
        out_df.columns = ['HEADER_TIME_STAMP']
        for col_name in col_names:
            out_df[col_name] = np.nan
        out_df = out_df.set_index('HEADER_TIME_STAMP')
    else:
        input_data = pd.read_csv(
            filepath, header=0, index_col=None, infer_datetime_format=True, parse_dates=[0])
        input_data = input_data.drop_duplicates(
            subset=['HEADER_TIME_STAMP'], keep='first')
        input_data = input_data.sort_values(by=['HEADER_TIME_STAMP'])
        input_data['HEADER_TIME_STAMP'] = input_data['HEADER_TIME_STAMP'] + \
            pd.Timedelta(1, unit='milliseconds')
        input_data = input_data.set_index('HEADER_TIME_STAMP')
        out_df = input_data.reindex(
            sample_ts, axis='index', method='nearest', tolerance=tolerance, limit=1)
        out_df.index.names = ['HEADER_TIME_STAMP']
    return out_df


def _data_to_annotation(hourly_df, sr=50):
    test = hourly_df.iloc[:, 0].copy(deep=True)
    test.loc[test.notna()] = 0
    test = test.fillna(1)
    edges = test.diff()
    if test.iloc[0] == 0:
        edges[0] = 0
    else:
        edges[0] = 1
    sts = hourly_df.loc[edges == 1].index.tolist()
    ets = hourly_df.loc[edges == -1].index.tolist()
    if len(sts) > len(ets):
        ets += [hourly_df.index[-1]]

    out_df = pd.DataFrame.from_dict({
        'START_TIME': sts,
        'STOP_TIME': ets,
        'PREDICTION': ["Missing"]*len(sts)
    })
    return out_df


def save_as_signaligner(df, output_path, file_type: FileType, **kwargs):
    if file_type == FileType.SENSOR:
        actigraph.save_as_actigraph(df, output_path, **kwargs)
    elif file_type == FileType.ANNOTATION:
        _save_annotation(df, output_path, **kwargs)


def _save_annotation(df, output_path, labelset, **kwargs):
    if 'START_TIME' not in df or 'STOP_TIME' not in df or ('LABEL_NAME' not in df and 'PREDICTION' not in df):
        raise ValueError('Input dataframe does not have proper columns.')
    if 'LABEL_NAME' in df:
        predictions = df['LABEL_NAME']
    elif 'PREDICTION' in df:
        predictions = df['PREDICTION']
    converted = {
        'START_TIME': df['START_TIME'].apply(format_time),
        'STOP_TIME': df['STOP_TIME'].apply(format_time),
        'PREDICTION': predictions,
        'SOURCE': ['Algo'] * df.shape[0],
        'LABELSET': [labelset] * df.shape[0]}
    converted = pd.DataFrame.from_dict(converted)
    converted.to_csv(output_path, **kwargs)



from .. import moment
import numpy as np
import pandas as pd
import datetime as dt
import tzlocal
import pytz
import time


class TestMoment:
    def test_moment(self):
        test0 = float(1580780607)
        test = test0
        assert moment.Moment(test)._posix == test0
        test = dt.datetime.fromtimestamp(test0)
        assert moment.Moment(test)._posix == test0
        test = np.datetime64(test)
        assert moment.Moment(test)._posix == test0
        test = pd.Timestamp.fromtimestamp(test0)
        assert moment.Moment(test)._posix == test0
        test = pd.Timestamp(test0, unit='s', tz=tzlocal.get_localzone())
        assert moment.Moment(test)._posix == test0

    def test_to(self):
        test0 = float(1580780607)
        test = moment.Moment(test0)
        assert test.to_unix_timestamp() == test0
        assert test.to_datetime() == dt.datetime.fromtimestamp(test0)
        assert test.to_datetime(
            tz=moment.Moment.get_utc_timezone()) == dt.datetime.fromtimestamp(test0, moment.Moment.get_utc_timezone())
        assert test.to_datetime(
            tz=moment.Moment.get_local_timezone()) == dt.datetime.fromtimestamp(test0, moment.Moment.get_local_timezone())
        assert test.to_datetime64() == np.datetime64(dt.datetime.fromtimestamp(test0))
        assert test.to_pandas_timestamp() == pd.Timestamp.fromtimestamp(test0)
        assert test.to_pandas_timestamp(tz=moment.Moment.get_utc_timezone(
        )) == pd.Timestamp(test0, unit='s', tz=moment.Moment.get_utc_timezone())
        assert test.to_string(fmt='%Y') == '2020'
        assert test.to_string(fmt='%Y%z') == '2020'
        assert test.to_string(
            fmt='%Y%z', tz=moment.Moment.get_utc_timezone()) == '2020+0000'

    def test_time_zones(self):
        assert moment.Moment.get_timezone('UTC') == pytz.utc

    def test_get_duration(self):
        ts = time.time()
        assert moment.Moment.get_duration(ts, ts + 1) == 1

        ts = time.time()
        a = [ts] * 10
        b = [ts + 10] * 10
        np.testing.assert_array_equal(
            moment.Moment.get_durations(a, b), 10)

    def test_get_sequence(self):
        st = time.time()
        result = moment.Moment.get_sequence(st, sr=1, N=2)
        assert moment.Moment.get_duration(result[0], result[1]) == 1
        result = moment.Moment.get_sequence(st, sr=100, N=2)
        assert moment.Moment.get_duration(result[0], result[1]) == 1 / 100.0
        result = moment.Moment.get_sequence(st, sr=1000, N=2)
        np.testing.assert_almost_equal(moment.Moment.get_duration(
            result[0], result[1]), 1 / 1000.0, decimal=2)
        result = moment.Moment.get_sequence(st, sr=1, N=2, format='posix')
        assert moment.Moment.get_duration(result[0], result[1]) == 1
        result = moment.Moment.get_sequence(st, sr=100, N=2, format='posix')
        np.testing.assert_almost_equal(moment.Moment.get_duration(
            result[0], result[1]), 1 / 100.0, decimal=2)
        result = moment.Moment.get_sequence(st, sr=1000, N=2, format='posix')
        np.testing.assert_almost_equal(moment.Moment.get_duration(
            result[0], result[1]), 1 / 1000.0, decimal=2)

        result = moment.Moment.get_sequence(
            st, sr=1, N=2, endpoint_as_extra=False, format='posix')
        assert moment.Moment.get_duration(result[0], result[1]) == 1
        assert len(result) == 2

    def test_seq_to_unix_timestamp(self):
        st0 = time.time()
        et = st0 + 1
        ts0 = np.arange(st0, et, step=1.0/50)
        result = moment.Moment.seq_to_unix_timestamp(ts0)
        np.testing.assert_array_equal(result, ts0)

        st = dt.datetime.fromtimestamp(st0)
        ts = []
        for n in range(0, 50):
            ts.append(st + dt.timedelta(milliseconds=1000.0/50 * n))
        result = moment.Moment.seq_to_unix_timestamp(ts)
        np.testing.assert_array_almost_equal(result, ts0, decimal=3)

        st = dt.datetime.fromtimestamp(st0)
        ts = []
        for n in range(0, 50):
            ts.append(st + dt.timedelta(milliseconds=1000.0/50 * n))
        ts = pd.to_datetime(ts).values
        result = moment.Moment.seq_to_unix_timestamp(ts)
        np.testing.assert_array_almost_equal(result, ts0, decimal=3)



"""
Utilities for developers.

1. Zip and release datasets to github.
2. Bump package versions.
3. Build packages and apps.
4. Control logger handlers.

Author: Qu Tang
Date: 01/30/2020
License: GNU v3
"""
import tarfile
import os
from loguru import logger
import sys
import subprocess
import pathlib
import shutil
import importlib
import pprint
import re

from . import mhealth_format as mh


def _print_extra_dep_warning(e):
    msg = (
        "Arus developer module requirements are not installed. Some functionality may not be working.\n\n"
        "To use the full functionality, please install the metawear extra packages as follows:\n\n"
        "  pip install arus[dev]\n\n"
    )
    print(str(e) + "\n\n" + msg)


def compress_dataset(source_dir, out_dir, out_name):
    os.makedirs(out_dir, exist_ok=True)
    if command_is_available('tar --version'):
        logger.info('Use tar to compress dataset...')
        output_path = os.path.join(out_dir, out_name)
        subprocess.run(
            f'tar --exclude=".git" --exclude="DerivedCrossParticipants" --exclude=".gitignore" -zcvf {output_path} -C {source_dir} *', shell=True)
    else:
        logger.info('Use Python tar module to compress dataset...')

        def exclude(tarinfo):
            if 'DerivedCrossParticipants' in tarinfo.name or '.git' in tarinfo.name or '.gitignore' in tarinfo.name:
                return None
            else:
                return tarinfo
        with tarfile.open(os.path.join(out_dir, out_name), "w:gz") as tar:
            tar.add(source_dir, arcname=os.path.basename(
                source_dir), filter=exclude)

        logger.info('Compression is completed.')


def command_is_available(cmd):
    status, output = subprocess.getstatusoutput(cmd + " --version")
    if status == 0:
        return True
    else:
        return False


def _find_current_version(root, name):
    package_module = importlib.import_module(name)
    if hasattr(package_module, '__version__'):
        return package_module.__version__
    if command_is_available('poetry') and os.path.exists(os.path.join(root, 'pyproject.toml')):
        return subprocess.getoutput('poetry version').split(' ')[1]


def _bump_news_version(new_version, dev_version=False):
    new_lines = []
    with open('news.md', 'r') as f:
        lines = f.readlines()
        if dev_version:
            new_lines = ['# ' + new_version + '\n\n'] + lines
        else:
            lines[0] = '# ' + new_version + '\n'
            new_lines = lines
    with open('news.md', 'w') as f:
        f.writelines(new_lines)


def bump_package_version(root, name, nver, dev=False):
    """Bump package version

    Args:
        root (str): The root path of the package
        name (str): The name of the package
        nver (str): The new version of the package
        dev (bool, optional): Whether it is a development version. Defaults to False.

    Returns:
        str or None: New version or None
    """
    try:
        deph = importlib.import_module('dephell_versioning')
    except ImportError as e:
        _print_extra_dep_warning(e)
        raise ImportError(e)
    cver = _find_current_version(root, name)
    if nver in ['major', 'minor', 'patch']:
        nver = deph.bump_version(
            version=cver, rule=nver, scheme='semver')
    if dev:
        nver = nver + '+9000'
    logger.info('new version is ' + nver)
    confirm = input("Confirm to continue [y/n]?")
    if confirm.lower() == 'y':
        if command_is_available('poetry') and os.path.exists(os.path.join(root, 'pyproject.toml')):
            logger.info('Bump poetry package version')
            subprocess.run("poetry version " + nver, shell=True, cwd=root)

        if os.path.exists(os.path.join(root, name, '__init__.py')):
            logger.info('Bump package version file')
            deph.bump_file(path=pathlib.Path('arus', '__init__.py'),
                           old=cver, new=nver)
        return nver
    else:
        return None


def commit_repo(message, repo_root=None):
    assert command_is_available('git')
    repo_root = repo_root or os.getcwd()
    logger.info('Commit with message: ' + message)
    subprocess.run('git add .', cwd=repo_root, shell=True)
    subprocess.run('git commit -m "{}"'.format(message),
                   cwd=repo_root, shell=True)


def tag_repo(version, message=None, repo_root=None):
    assert command_is_available('git')
    repo_root = repo_root or os.getcwd()
    message = message or "Tag with version {}".format(version)
    logger.info(
        'Tag repo with version {} and message: {}'.format(version, message))
    subprocess.run(
        f'git tag -a v{version} -m "{message}"', cwd=repo_root, shell=True)


def push_repo(branch='master', repo_root=None):
    assert command_is_available('git')
    repo_root = repo_root or os.getcwd()
    subprocess.run('git push origin {}'.format(
        branch), cwd=repo_root, shell=True)


def push_tag(version, repo_root=None):
    assert command_is_available('git')
    repo_root = repo_root or os.getcwd()
    logger.info('Push tag v{}'.format(version))
    subprocess.run('git push origin v' + version, cwd=repo_root, shell=True)


def build_arus_app(root, app_name, version):
    logger.info('Build {}'.format(app_name))
    shutil.rmtree('./apps/{}/build'.format(app_name), ignore_errors=True)
    shutil.rmtree('./apps/{}/dist'.format(app_name), ignore_errors=True)

    cwd = os.path.join(root, 'apps', app_name)
    subprocess.run("pyinstaller main.spec", shell=True, cwd=cwd)

    app_path = os.path.join(os.getcwd(), 'apps', app_name, 'releases',
                            app_name + '_' + version + '.zip')

    os.makedirs(os.path.dirname(app_path), exist_ok=True)
    cwd = os.path.join(os.getcwd(), 'apps', app_name, 'dist', app_name)

    subprocess.run("zip -r " + app_path +
                   " ./", shell=True, cwd=cwd)
    shutil.rmtree('./apps/{}/build'.format(app_name), ignore_errors=True)


def _copy_files_for_website():
    src_files = [
        "README.md",
        "LICENSE.md",
        "CODE_OF_CONDUCT.md"
    ]
    dest_files = list(map(lambda f: os.path.join('docs', f), src_files))
    for src_file, dest_file in zip(src_files, dest_files):
        src_file = os.path.join('.', src_file)
        dest_file = os.path.join('.', dest_file)
        if os.path.exists(dest_file):
            os.remove(dest_file)
        logger.info(f'Copy {os.path.basename(src_file)} to docs folder')
        shutil.copyfile(src_file, dest_file)


def dev_website():
    _copy_files_for_website()
    generate_changelogs()
    subprocess.run("mkdocs serve -v", shell=True)


def build_website():
    _copy_files_for_website()
    generate_changelogs()
    subprocess.run("mkdocs gh-deploy -v", shell=True)


def get_git_tags():
    try:
        semver = importlib.import_module('semver')
    except ImportError as e:
        _print_extra_dep_warning(e)
        raise ImportError(e)
    tags = subprocess.check_output(
        " ".join(['git', 'tag']), shell=True, encoding='utf-8', text=True, universal_newlines=True)
    tags = tags.split('\n')
    tags = list(map(lambda tag: tag.strip('v'), filter(
        lambda tag: tag.startswith('v'), tags)))
    tags = sorted(tags, key=semver.VersionInfo.parse, reverse=True)
    return tags


def generate_changelogs(by_version=True):
    tags = get_git_tags()
    stop_tag = 'HEAD'
    if by_version:
        for start_tag in tags:
            start_tag = f'v{start_tag}'
            changelogs = subprocess.check_output(
                " ".join(['git', 'log', f'{start_tag}...{stop_tag}', '--pretty=format:%s-%h']), shell=True, encoding='utf-8',
                text=True, universal_newlines=True).split('\n')
            parsed = parse_changelogs(changelogs)
            write_changelog_to_file(parsed, start_tag, stop_tag)
            stop_tag = start_tag
    write_changelog_index_to_file(tags)
    return True


def parse_changelogs(changelogs):
    parsed = {}
    category_map = {
        'api': "API changes",
        'feat': "Features",
        'refactor': "Refactors",
        'fix': "Bug fixes",
        'test': "Test cases"
    }
    for changelog in changelogs:
        tokens = changelog.split('-')
        if len(tokens) < 2:
            logger.warning(f'Commit {changelog} is invalid, ignore it')
            continue
        msg = tokens[0]
        commit = tokens[1]
        category, scope, title = parse_conventional_commit(msg)
        category = category.lower()
        if category == "" or category not in category_map:
            logger.warning(
                f'Commit {commit} is not a conventional commit, ignore it')
            continue
        category = category_map[category]
        if category in parsed:
            parsed[category].append(
                {'scope': scope, 'title': title, 'commit': commit})
        else:
            parsed[category] = [
                {'scope': scope, 'title': title, 'commit': commit}]
    return parsed


def write_changelog_index_to_file(tags):
    item_template = "* [{tag}]({tag}/)\n"
    with open(os.path.join('.', 'docs', 'changelogs', 'index.md'), 'w', encoding='utf-8') as f:
        f.write(f'# Changelogs (Latest stable version: {tags[0]})\n\n')
        f.write(item_template.format(tag='dev'))
        for tag in tags:
            tag = f'v{tag}'
            f.write(item_template.format(tag=tag))


def write_changelog_to_file(parsed_changelogs, start_tag, stop_tag):
    template = "## {category}"

    item_template = "* {title} [{commit}](https://github.com/qutang/arus/commit/{commit}) (**{scope}**)"
    item_template2 = "* {title} [{commit}](https://github.com/qutang/arus/commit/{commit})"

    category_order = ['API changes',
                      'Features',
                      'Bug fixes',
                      'Refactors',
                      'Test cases']

    if stop_tag == 'HEAD':
        stop_tag = 'dev'

    os.makedirs(os.path.join('.', 'docs', 'changelogs'), exist_ok=True)
    with open(os.path.join('.', 'docs', 'changelogs', f'{stop_tag}.md'), 'w', encoding='utf-8') as f:
        f.write(f'# {stop_tag.upper()}\n')
        for category in category_order:
            if category not in parsed_changelogs:
                continue
            f.write(template.format(category=category))
            f.write('\n')
            for item in parsed_changelogs[category]:
                if item['scope'] == '':
                    f.write(item_template2.format(
                        title=item['title'], commit=item['commit']))
                else:
                    f.write(item_template.format(
                        scope=item['scope'], title=item['title'], commit=item['commit']))
                f.write('\n')


def parse_conventional_commit(msg):
    tokens = msg.split(':')
    if len(tokens) == 1:
        return "", "", ""
    category_scope = tokens[0]
    title = tokens[1].strip(' ')
    sub_tokens = re.findall(r'\w+', category_scope)
    category = sub_tokens[0]
    if len(sub_tokens) > 1:
        scope = sub_tokens[1]
    else:
        scope = ""
    return category, scope, title


def logging_dict(data):
    info = pprint.pformat(data, width=1)
    logger.info(info)


def logging_st_and_et(st, et):
    st_str = st.strftime('%Y-%m-%d %H:%M:%S.%f')
    et_str = et.strftime('%Y-%m-%d %H:%M:%S.%f')
    logger.info("{} - {}".format(st_str, et_str))



import queue
import threading
import time
import enum
from loguru import logger
import typing

from . import operator
from . import node


class Pipeline(operator.Operator):
    """
    The base class for data processing pipeline for HAR.

    Stream class is an abstraction of any data source that can be loaded into memory in arbitrary chunk size either asynchronously (currently only support threading) or synchronously.
    """

    def __init__(self, *streams, synchronizer, processor,  name: typing.Optional[str] = 'default-pipeline'):
        """[summary]

        Arguments:
            synchronizer {[type]} -- [description]
            processor {[type]} -- [description]

        Keyword Arguments:
            name {typing.Optional[str]} -- [description] (default: {'default-pipeline'})
        """
        super().__init__()
        self._name = name
        self._streams = [node.Node(op=stream, t=node.Node.Type.INPUT,
                                   name=self._name + stream._name) for stream in streams]
        self._synchronizer = node.Node(op=synchronizer, t=node.Node.Type.PIPE,
                                       name=self._name + '-synchronizer')
        self._processor = node.Node(op=processor, t=node.Node.Type.PIPE,
                                    name=self._name + '-processor')

    def run(self, *, values=None, src=None, context={}):
        logger.info('Stream is starting.')
        for stream in self._streams:
            stream.start()
        self._synchronizer.start()
        self._processor.start()
        logger.info('Stream started.')

    def start(self, start_time: "str, datetime, numpy.datetime64, pandas.Timestamp" = None):
        """Method to start loading data from the provided data source.

        Arguments:
            start_time: The reference time for segmentation.

        Note:
            `start_time` is used to sync between multiple streams. If it is `None`, the default value would be extracted from the first sample of the loaded data.
        """
        self._context['ref_start_time'] = start_time
        self.run()

    def stop(self):
        super().stop()
        """Stop the loading process."""
        logger.info('Stream is stopping.')
        self._processor.stop()
        logger.info('Processor thread stopped.')
        time.sleep(0.1)
        self._synchronizer.stop()
        logger.info('Synchronizer thread stopped.')
        time.sleep(0.1)
        for stream in self._streams:
            stream.stop()
        logger.info('Stream threads stopped.')
        logger.info('Stream stopped.')

    def shutdown(self):
        self._processor.get_op().shutdown()

    def get_result(self):
        while True:
            if self._stop:
                break
            for stream in self._streams:
                data = next(stream.produce())
                self._synchronizer.consume(data)
                data = next(self._synchronizer.produce())
                self._processor.consume(data)
                data = next(self._processor.produce())
                if data.signal == node.Node.Signal.WAIT:
                    pass
                elif data.signal == node.Node.Signal.DATA:
                    yield data.values, data.context
                else:
                    break
        yield None, {}
